// The BSP structure
// altered for unmap.Vmex


package unmap;
import java.text.*;
import java.nio.*;
import java.util.*;

public class BSP
{
    //header
    
    int ident;
    int version;
    Lump lump[];
    int maprev;
    
    //selected lump data
    
    boolean no_decomp = false;
    
    //vertexes
    
    int verts;
    float x[], y[], z[];   //coordinates
    boolean o[];	//true if vertex has been shifted from origin
    
    //edges
    
    int edges;
    int vi1[], vi2[];		// start and end coordinate index
    byte eim[];			// edge marker
    boolean t[];
    
    //surfedges
    int surfedges;
    int sedge[];		//
    
    //faces
    int faces;
    Face face[];
    
    //models
    int models;
    Model mod[];
    
    //entities
    int entities;
    Entity ent[];
    
    //original faces;
    int ofaces;
    Face oface[];
    Wind owind[];   // used in brush-model matching
    
    int nodes;
    Node node[];
    
    int leaves;
    Leaf leaf[];
    
    int lfaces;
    short lface[];
    
    int lbrushes;
    short lbrush[];
    
    int texinfos;
    Texinfo tx[];
    
    int texdatas;
    Texdata td[];
    
    int tdsts;		// texdata string table
    int tdst[];		// index of string in tdsd
    
    int tdsds;		// texdata string data
    byte [] tdsd;	// the string byte array
    
    String [] texname;	// the string array
    int [] texenv;      // the envmap associated with this texture
    
    int propstatics;
    Pstatic ps[];
    int psnames;
    String [] psname;
    
    int glumps;
    GLump [] gl;
    
    int cubemaps;
    Cubemap [] cm;
    
    int dispinfos;
    Dispinfo [] dinfo;
    
    int dispverts;
    Dispvert [] dv;
    
    int disptris;
    short [] dtri;
    
    int planes;
    Plane [] pl;
    
    int brushes;
    Brush [] br;
    
    int brushsides;
    Brushside [] bsd;
    Wind [] bwind;      // used in brush-model matching
    
    
    
    public void loadmap(ByteBuffer b)
    {
        
        //Load the BSP header
        System.out.println("Header");
        loadheader(b);
        
        // VERTEXES 3
        System.out.print("Vertexes ");
        loadverts(b);
        
        // EDGES 12
        System.out.print("Edges ");
        loadedges(b);
        
        
        //FACES 7
        System.out.print("Faces ");
        loadfaces(b);
        
        //ORIGINAL FACES 27
        System.out.print("Orig faces ");
        loadorigfaces(b);
        
        //MODELS 14
        System.out.print("Models ");
        loadmodels(b);
        
        //SURFEDGE 13
        System.out.print("Surfedges ");
        loadsurfedges(b);
        
        //MARK MODEL EDGES
        System.out.println("Marks");
        markedges();
        
        //TEXINFOS 6
        System.out.print("Texinfo ");
        loadtexinfos(b);
        
        //TEXDATA 2
        System.out.print("Texdata ");
        loadtexdatas(b);
        
        //TEXDATASTRINGDATA 43
        System.out.println("Texdata string data");
        loadtdsd(b);
        
        //TEXDATASTRINGTABLE 44
        System.out.print("Texdata string table ");
        loadtdst(b);
        
        //ENTITIES 0
        System.out.print("Entities ");
        loadentities(b);
        
        transmodels();  // no origin move
        
        //PROP STATICS (gamelump 35)
        loadgamelumpheader(b);
        System.out.print("Prop_statics ");
        loadstatics(b);
        
        //CUBEMAPS 42
        System.out.print("Cubemaps ");
        loadcubemaps(b);
        
        System.out.print("Planes ");
        loadplanes(b);
        
        System.out.print("Brushes ");
        loadbrushes(b);
        
        System.out.print("Brushsides ");
        loadbrushsides(b);
        
        System.out.print("Dispinfo ");
        loaddispinfo(b);
        
        System.out.print("Dispverts ");
        loaddispverts(b);
        
        System.out.print("Disptris ");
        loaddisptris(b);
        
        //NODES 5
        System.out.print("Nodes ");
        loadnodes(b);
        
        //LEAFS 10
        System.out.print("Leaves ");
        loadleaves(b);
        
        System.out.print("Leaf faces ");
        loadlfaces(b);
        
        System.out.print("Leaf brushes ");
        loadlbrushes(b);
        
        
        
    }
    
    public void loadheader(ByteBuffer b)
    {
        //Read the BSP Header
        
        b.position(0);
        
        ident = b.getInt();
        int idbsp = ('P'<<24)+('S'<<16)+('B'<<8)+'V';
        version = b.getInt();
        
        System.out.println("Ident: "+ident);
        System.out.println("Version: "+version);
        
        if(ident != idbsp || version!=19)
        {
            System.out.println("Unknown map file ident or version!");
            System.exit(1);
        }
        
        lump = new Lump[64];
        
        for(int i=0;i<64;i++)
        {
            lump[i] = new Lump();
            
            lump[i].ofs=b.getInt();
            lump[i].len=b.getInt();
            lump[i].vers=b.getInt();
            lump[i].fourCC=b.getInt();
            
            if(lump[i].len>0)
            {
                System.out.print("Lump "+i+": ");
                System.out.print(lump[i].ofs+", "+lump[i].len+", "+lump[i].vers+", "+lump[i].fourCC);
                System.out.println(" "+Lump.name(i)+"  "+lump[i].len/Lump.size(i)+(Lump.size(i)==1?" bytes":""));
            }
        }
        
        maprev = b.getInt();
        System.out.println("MapRev: "+maprev);
        
    }
    
    public int sefix(int edge)
    {
        int iedge = sedge[edge];
        if(iedge<0)
        {
            iedge = -iedge;
        }
        return iedge;
    }
    
    public void loadplanes(ByteBuffer b)      // PLANES 1
    {
        b.position(lump[1].ofs);
        
        planes = lump[1].len/Lump.size(1);
        
        pl = new Plane[planes];
        
        for(int i=0;i<planes;i++)
        {
            pl[i] = new Plane();
            
            pl[i].nx = b.getFloat();
            pl[i].ny = b.getFloat();
            pl[i].nz = b.getFloat();
            pl[i].dist = b.getFloat();
            pl[i].type = b.getInt();
        }
        System.out.println(planes);
        
    }
    
    public void loadbrushes(ByteBuffer b)      // BRUSHES 18
    {
        b.position(lump[18].ofs);
        
        brushes = lump[18].len/Lump.size(18);
        
        br = new Brush[brushes];
        
        for(int i=0;i<brushes;i++)
        {
            br[i] = new Brush();
            
            br[i].fstside = b.getInt();
            br[i].numside = b.getInt();
            br[i].contents = b.getInt();
        }
        System.out.println(brushes);
    }
    
    public void loadbrushsides(ByteBuffer b)      // BRUSHSIDES 19
    {
        b.position(lump[19].ofs);
        
        brushsides = lump[19].len/Lump.size(19);
        
        bsd = new Brushside[brushsides];
        
        for(int i=0;i<brushsides;i++)
        {
            bsd[i] = new Brushside();
            
            bsd[i].pnum = (int) 0xFFFF & b.getShort();     //
            bsd[i].texinfo = b.getShort();
            bsd[i].dispinfo = b.getShort();
            bsd[i].bevel = b.getShort();
        }
        System.out.println(brushsides);
    }
    
    
    public void loadverts(ByteBuffer b)      // VERTEXES 3
    {
        b.position(lump[3].ofs);
        
        verts = lump[3].len/Lump.size(3);
        
        x = new float[verts];
        y = new float[verts];
        z = new float[verts];
        o = new boolean[verts];
        
        float M = 1e6f;
        float xmin = M, xmax = -M, ymin = M, ymax = -M, zmin = M, zmax = -M;
        
        for(int i=0;i<verts;i++)
        {
            x[i]=b.getFloat(); xmin=Math.min(x[i],xmin); xmax=Math.max(x[i],xmax);
            y[i]=b.getFloat(); ymin=Math.min(y[i],ymin); ymax=Math.max(y[i],ymax);
            z[i]=b.getFloat(); zmin=Math.min(z[i],zmin); zmax=Math.max(z[i],zmax);
            o[i]=false;
            
            //	System.out.println(x[i]+", "+y[i]+", "+z[i]);
            
        }
        System.out.println(verts);
        System.out.println("Min/Max: ("+xmin+", "+ymin+", "+zmin+") - ("+xmax+", "+ymax+", "+zmax+")");
        
    }
    
    public void loadedges(ByteBuffer b)
    {
        // EDGES 12
        b.position(lump[12].ofs);
        
        edges = lump[12].len/Lump.size(12);
        
        vi1 = new int[edges];
        vi2 = new int[edges];
        eim = new byte[edges];
        t = new boolean[edges];
        
        int vmax = 0;
        
        for(int i=0;i<edges;i++)
        {
            
            vi1[i]=b.getShort();	vmax = Math.max(vi1[i],vmax);
            vi2[i]=b.getShort(); 	vmax = Math.max(vi2[i],vmax);
            eim[i]=0;
            t[i]=false;
        }
        
        if(vmax>verts)
        {
            System.out.println("Vertex index to high "+vmax+" > "+verts);
            return;
        }
        System.out.println(edges);
    }
    
    public void loadfaces(ByteBuffer b)
    {
        //FACES 7
        
        b.position(lump[7].ofs);
        
        faces = lump[7].len/Lump.size(7);
        
        face = new Face[faces];
        
        for(int i=0;i<faces;i++)
        {
            face[i] = new Face();
            
            face[i].pnum = (int) 0xFFFF & b.getShort(); //planenum
            face[i].side = b.get();             //side
            face[i].onnode = b.get();           //onnode
            face[i].fstedge = b.getInt();      	//1st edge
            face[i].numedge = b.getShort();     //num edges
            face[i].texinfo = b.getShort();     //texinfo
            face[i].dispinfo = b.getShort();     //dispinfo
            b.getShort();     //surfvolume
            
            b.get();          //styles[4]
            b.get();
            b.get();
            b.get();
            
            b.getInt();      //lightofs
            face[i].area = b.getFloat();    //area
            b.getInt();      //LightmapTexture[4]
            b.getInt();
            b.getInt();
            b.getInt();
            face[i].orig = b.getInt();      //origface
            b.getShort();    //numprimitives
            b.getShort();    //firstprimitive
            face[i].smooth = b.getInt();      //smoothgroups
        }
        System.out.println(faces);
        
    }
    
    public void loadorigfaces(ByteBuffer b)
    {
        //ORIGINAL FACES 27
        
        b.position(lump[27].ofs);
        
        ofaces = lump[27].len/Lump.size(27);
        
        oface = new Face[ofaces];
        
        for(int i=0;i<ofaces;i++)
        {
            oface[i] = new Face();
            
            oface[i].pnum = (int) 0xFFFF & b.getShort(); //planenum
            oface[i].side = b.get();             //side
            oface[i].onnode = b.get();           //onnode
            oface[i].fstedge = b.getInt();      	//1st edge
            oface[i].numedge = b.getShort();     	//num edges
            oface[i].texinfo = b.getShort();      //texinfo
            
            oface[i].dispinfo = b.getShort();     //dispinfo
            
            b.getShort();        //surfvolume
            
            b.get();            //styles[4]
            b.get();
            b.get();
            b.get();
            
            b.getInt();       //lightofs
            b.getFloat();       //area
            oface[i].area = 0.0f;   // it's zero anyway, but just in case
            b.getInt();         //LightmapTexture[4]
            b.getInt();
            b.getInt();
            b.getInt();
            oface[i].orig = b.getInt();       //origface
            b.getShort();     //numprimitives
            b.getShort();    //firstprimitive
            oface[i].smooth = b.getInt();      //smoothgroups
        }
        System.out.println(ofaces);
    }
    
    public void loadmodels(ByteBuffer b)
    {
        //MODELS 14
        
        b.position(lump[14].ofs);
        models = lump[14].len/Lump.size(14);
        
        mod = new Model[models];
        
        for(int i=0;i<models;i++)
        {
            mod[i] = new Model();
            
            b.getFloat(); //mxl
            b.getFloat(); //myl
            b.getFloat(); //mzl
            b.getFloat(); //mxh
            b.getFloat(); //myh
            b.getFloat(); //mzh
            b.getFloat(); //mox
            b.getFloat(); //moy
            b.getFloat(); //moz
            mod[i].headnode = b.getInt(); //headnode
            mod[i].fstface = b.getInt();
            mod[i].numface = b.getInt();
            //System.out.print("Model "+i+" o:"+mox+","+moy+","+moz+"\t");
        }
        System.out.println(models);
    }
    
    public void loadsurfedges(ByteBuffer b)
    {
        //SURFEDGE 13
        
        b.position(lump[13].ofs);
        surfedges = lump[13].len/Lump.size(13);
        
        sedge = new int[surfedges];
        
        for(int i=0;i<surfedges;i++)
        {
            sedge[i] = b.getInt();
            
        }
        System.out.println(surfedges);
    }
    
    public void loadgamelumpheader(ByteBuffer b)
    {
        //GAME LUMP 35
        
        b.position(lump[35].ofs);
        
        glumps = b.getInt();	// GL header count
        
        gl = new GLump[glumps];
        
        System.out.println(glumps+" game lumps");
        
        for(int i=0;i<glumps;i++)
        {
            gl[i] = new GLump();
            
            gl[i].id = b.getInt();
            gl[i].flags = b.getShort();
            gl[i].vers =  b.getShort();
            
            gl[i].ofs = b.getInt();
            gl[i].len = b.getInt();
            
            //			char []id = new char[4];
            //			id[0] = (char)(0xFF & gl[i].id >> 24);
            //			id[1] = (char)(0xFF & gl[i].id >> 16);
            //			id[2] = (char)(0xFF & gl[i].id >> 8);
            //			id[3] = (char)(0xFF & gl[i].id);
            
            //			String sid = new String(id);
            //			System.out.println("GLID: "+gl[i].id+"  "+sid);
            //			System.out.println("F: "+gl[i].flags+" V: "+gl[i].vers);
            //			System.out.println("O: "+gl[i].ofs+" L: "+gl[i].len);
            
        }
    }
    
    public void loadstatics(ByteBuffer b)
    {
        // STATIC PROPS (Gamelump 35)
        int spid = -1;
        for(int i=0;i<glumps;i++)
        {
            if(gl[i].id == 1936749168 )		// 'sprp'
                spid = i;
        }
        if(spid<0)	return;		// no static prop gamelump
        b.position(gl[spid].ofs);	// offset of sprp data
        int	pslen = gl[spid].len;
        
        
/*		StringBuffer text = new StringBuffer();
 
                System.out.println();
                for(int i=0;i<pslen;i++)
                {
                        int x = (int)0xFF & b.get();
                        if(x<16)	System.out.print("0");
                        System.out.print(Integer.toHexString(x)+" ");
                        if(x>31)	text.append((char)x); else text.append("?");
                        if(i%16 == 15)
                        {
                                System.out.println("     "+text.toString());
                                text.delete(0,text.length());
                        }
                }
 
                b.position(gl[spid].ofs);  */
        
        psnames = b.getInt();		// number of entries in dictionary
        psname = new String[psnames];
        
        for(int i=0;i<psnames;i++)
        {
            psname[i] = readstr(b,128);		// 128 bytes in each name
        }
        
        int propleaves = b.getInt();	// number of leafprops - not important
        
        for(int i=0;i<propleaves;i++)
            b.getShort();		// skip over them
        
        propstatics = b.getInt();
        
        ps = new Pstatic[propstatics];
        
        for(int i=0;i<propstatics;i++)
        {
            ps[i] = new Pstatic();
            
            ps[i].ox = b.getFloat();	// origin vector
            ps[i].oy = b.getFloat();
            ps[i].oz = b.getFloat();
            
            ps[i].ax = b.getFloat();	// angles pyr
            ps[i].ay = b.getFloat();
            ps[i].az = b.getFloat();
            
            ps[i].type = b.getShort();	// prop_static type
            
            b.getShort();			// first leaf
            b.getShort();			// leaf count
            ps[i].solid = b.get();
            ps[i].flags = b.get();
            ps[i].skin = b.getInt();
            ps[i].fademin = b.getFloat();
            ps[i].fademax = b.getFloat();
            
            b.getFloat();			// lighting vec
            b.getFloat();
            b.getFloat();
            if(gl[spid].vers==5) b.getInt();		//pad if version 4
            
        }
        System.out.println(propstatics);
    }
    
    public void loadcubemaps(ByteBuffer b)
    {   //CUBEMAPS 42
        
        
        b.position(lump[42].ofs);
        cubemaps = lump[42].len/Lump.size(42);
        
        cm = new Cubemap[cubemaps];
        
        for(int i=0;i<cubemaps;i++)
        {
            cm[i] = new Cubemap();
            
            cm[i].x = b.getInt();
            cm[i].y = b.getInt();
            cm[i].z = b.getInt();
            
            cm[i].size = (byte)b.getInt();      // is a padded byte
            
        }
        System.out.println(cubemaps);
    }
    
    public void loaddispinfo(ByteBuffer b)
    {
        //DISPINFO 26
        b.position(lump[26].ofs);
        dispinfos = lump[26].len/Lump.size(26);
        
        dinfo = new Dispinfo[dispinfos];
        
        for(int i=0;i<dispinfos;i++)
        {
            Dispinfo di=new Dispinfo();
            
            di.sx = b.getFloat();
            di.sy = b.getFloat();
            di.sz = b.getFloat();
            di.ivert = b.getInt();
            di.itri = b.getInt();
            di.power = b.getInt();
            di.smoothangle = b.getFloat();
            di.contents = b.getInt();
            di.face = b.getShort();
            b.getShort();                       // alignment pad
            b.getInt();
            b.getInt();
            for(int j=0;j<92;j++)  b.get();        // pad 92 bytes
            for(int j=0;j<10;j++)   di.allow[j] = b.getInt();  // allowedverts
            
            dinfo[i] = di;
        }
        System.out.println(dispinfos);
    }
    
    public void loaddispverts(ByteBuffer b)
    {
        //DISPVERTS 33
        b.position(lump[33].ofs);
        dispverts = lump[33].len/Lump.size(33);
        
        dv = new Dispvert[dispverts];
        
        for(int i=0;i<dispverts;i++)
        {
            dv[i] =new Dispvert();
            
            dv[i].x = b.getFloat();
            dv[i].y = b.getFloat();
            dv[i].z = b.getFloat();
            
            dv[i].dist = b.getFloat();
            dv[i].alpha = b.getFloat();
            
            
        }
        System.out.println(dispverts);
    }
    
    public void loaddisptris(ByteBuffer b)
    {
        //DISPVERTS 48
        b.position(lump[48].ofs);
        disptris = lump[48].len/Lump.size(48);
        
        dtri = new short[disptris];
        
        for(int i=0;i<disptris;i++)
        {
            dtri[i] = b.getShort();
        }
        System.out.println(disptris);
    }
    
    public void loadtexinfos(ByteBuffer b)
    {
        //TEXINFO 6
        b.position(lump[6].ofs);
        texinfos = lump[6].len/Lump.size(6);
        
        tx = new Texinfo[texinfos];
        
        for(int i=0;i<texinfos;i++)
        {
            tx[i] = new Texinfo();
            
            for(int j=0;j<2;j++)
            {
                for(int k=0;k<4;k++)
                    tx[i].tv[j][k] = b.getFloat();	//tex vecs texelspwu [2][4]
            }
            
            for(int j=0;j<8;j++)	b.getFloat();		//lightmaps vecs luxelspwu [2][4]
            
            tx[i].flags = b.getInt();	//flags
            tx[i].texdata = b.getInt();	//texdata
        }
        System.out.println(texinfos);
    }
    
    public void loadtexdatas(ByteBuffer b)
    {
        //TEXDATA 2
        b.position(lump[2].ofs);
        texdatas = lump[2].len/Lump.size(2);
        
        td = new Texdata[texdatas];
        
        for(int i=0;i<texdatas;i++)
        {
            td[i] = new Texdata();
            
            b.getFloat();	b.getFloat();	b.getFloat();	//reflectivity vector
            td[i].nameid = b.getInt();	//name string ID
            b.getInt();	b.getInt();		// width, heigh
            b.getInt();	b.getInt();		// view width, heigh
        }
        System.out.println(texdatas);
    }
    
    public void loadtdsd(ByteBuffer b)
    {
        //TEXDATA STRING DATA 43
        b.position(lump[43].ofs);
        tdsds = lump[43].len/Lump.size(43);
        
        tdsd = new byte[tdsds];
        b.get(tdsd);		// the byte array
    }
    
    public void loadtdst(ByteBuffer b)
    {
        //TEXDATA STRING TABLE 44
        b.position(lump[44].ofs);
        tdsts = lump[44].len/Lump.size(44);
        
        tdst = new int[tdsts];
        texname = new String[tdsts];    // the texture names
        texenv = new int[tdsts];        // the texture envmaps
        
        for(int i=0;i<tdsts;i++)
        {
            int ix = tdst[i] = b.getInt();	// index into tdsd
            
            for(int j=ix;j<tdsds;j++)		// scan from ix forward
            {
                if(tdsd[j]==0)			// looking for null-terminated string
                {
                    texname[i] = new String(tdsd, ix, j-ix);	// construct string from subarray
                    texenv[i] = -1;     // no envmap yet
                    break;
                }
            }
        }
        System.out.println(tdsts);
    }
    
    
    
    public void loadentities(ByteBuffer b)		// load entities
    {
        //ENTITIES 0
        
        //first run through to find number of entities
        b.position(lump[0].ofs);
        
        int endent = lump[0].ofs+lump[0].len;
        int numents = 0;
        String line;
        
        //quick and dirty - count the "}"'s
        while(b.position()<endent)
        {
            line=readline(b,endent);
            if(b.position()>endent) break;
            if(line.equals("}"))	numents++;
        }
        entities = numents;
        b.position(lump[0].ofs);		//reset to beginning of entity lump
        ent = new Entity[entities];		// the entity array
        Entity nent = new Entity();		// prevent uninitialized
        boolean eoe=false, eoel=false;
        
        numents = 0;
        
        allread:
            
            while(!eoel)
            {
                while(!eoe)		//parse each entity
                {
                    line = readline(b, endent);
                    //System.out.println(line);
                    if(b.position()>endent) //end of entity lump
                    {
                        eoel=true;
                        break allread;
                    }
                    if(line.equals("{"))   //start of entity
                    {
                        
                        nent = new Entity();	// new temporary entity to load
                        nent.index = numents;	// position in the array
                        nent.model = 0;
                        nent.ox = 0.0f;		// default origin
                        nent.oy = 0.0f;
                        nent.oz = 0.0f;
                        nent.mark = false;	// marker for tree building
                        //System.out.println("SOE");
                        continue;
                    }
                    if(line.equals("}"))//end of entity
                    {
                        //System.out.println("EOE");
                        eoe=true; continue;
                    }
                    String [] keyval = line.split("\"");
                    
                    if(keyval.length==4)
                    {
                        String ckey = keyval[1];	// find key/value pair
                        String cval = keyval[3];
                        //System.out.println(ckey+" "+cval);
                        
                        nent.key.add(ckey);		// store all key/values
                        nent.value.add(cval);
                        nent.link.add(null);
                        
                        // parse known entities
                        if(ckey.equals("no_decomp"))
                            no_decomp = true;            // protected map
                        if(ckey.equals("classname"))
                            nent.classname = cval;
                        else if(ckey.equals("targetname"))
                            nent.targetname = cval;
                        else if(ckey.equals("parentname"))
                            nent.parentname = cval;
                        else if(ckey.equals("model"))
                        {
                            if(cval.startsWith("*"))
                                nent.model = Integer.parseInt(cval.substring(1)); // brush model #
                            else
                                nent.model = -1;    // prop model
                        }
                        else if(ckey.equals("origin"))
                        {
                            String [] costr = cval.split(" "); // parse coords
                            nent.ox = Float.parseFloat(costr[0]);
                            nent.oy = Float.parseFloat(costr[1]);
                            nent.oz = Float.parseFloat(costr[2]);
                        }
                        else if(nent.con<0)		// try to gauge start of connections block
                        {					// not found yet
                            
                            String [] constr = cval.split(",");	// split value at comma
                            if(constr.length == 5)		// five fields, probably a connection
                            {
                                if(constr[4].equals("1") || constr[4].equals("-1"))	// last field is 1 or -1
                                {				// almost certainly a connection
                                    nent.con = nent.key.size() - 1;	// this kv is start of connection block
                                }
                            }
                        }
                    }
                }
                if(nent.parentname != null && nent.targetname !=null)
                {
                    if(nent.parentname.equals("tlight") && nent.targetname.charAt(0)=='q')
                    {
                        int q=Integer.parseInt(nent.targetname.substring(1));
                        
                        System.out.println("Unknown entity lump format: 0!="+(q/2));
                        System.exit(1);
                    }
                }
                
                ent[numents] = nent;	// assign the loaded entity to the array
                
                eoe=false;
                //System.out.println("Entity: "+numents);
                numents++;
                
            }
            
            if(numents != entities)		// check that quick count and loaded count of entities match
                System.out.println("Error! Entity count mismatch "+numents+" != "+entities);
            System.out.println(entities);
            
            // no need for entity links, data tree, etc.
            
            
            
    }
    
    public void loadnodes(ByteBuffer b)
    {
        //NODES 5
        
        b.position(lump[5].ofs);
        nodes = lump[5].len/Lump.size(5);
        
        node = new Node[nodes];
        
        for(int i=0;i<nodes;i++)
        {
            node[i] = new Node();
            
            node[i].pnum = b.getInt();
            node[i].child[0] = b.getInt();
            node[i].child[1] = b.getInt();
            node[i].mins[0] = b.getShort();
            node[i].mins[1] = b.getShort();
            node[i].mins[2] = b.getShort();
            node[i].maxs[0] = b.getShort();
            node[i].maxs[1] = b.getShort();
            node[i].maxs[2] = b.getShort();
            node[i].fstface = b.getShort();
            node[i].numface = b.getShort();
            node[i].area = b.getShort();
            b.getShort(); //padding
            //System.out.println("Node "+i+" pnum "+node[i].pnum+" c "+node[i].child[0]+","+node[i].child[1]+
            //    " mins "+node[i].mins[0]+","+node[i].mins[1]+","+node[i].mins[2]+
            //    " maxs "+node[i].maxs[0]+","+node[i].maxs[1]+","+node[i].maxs[2]+
            //    " ff "+node[i].fstface+" nf "+node[i].numface+" a "+node[i].area);
            
        }
        System.out.println(nodes);
    }
    
    public void loadleaves(ByteBuffer b)
    {
        //LEAFS 10
        
        b.position(lump[10].ofs);
        leaves = lump[10].len/Lump.size(10);
        
        leaf = new Leaf[leaves];
        
        for(int i=0;i<leaves;i++)
        {
            
            leaf[i] = new Leaf();
            
            leaf[i].contents = b.getInt();
            leaf[i].cluster  = b.getShort();
            leaf[i].area_flags = b.getShort();
            leaf[i].mins[0] = b.getShort();
            leaf[i].mins[1] = b.getShort();
            leaf[i].mins[2] = b.getShort();
            leaf[i].maxs[0] = b.getShort();
            leaf[i].maxs[1] = b.getShort();
            leaf[i].maxs[2] = b.getShort();
            
            leaf[i].fstleafface = b.getShort();
            leaf[i].numleafface = b.getShort();
            
            leaf[i].fstleafbrush = b.getShort();
            leaf[i].numleafbrush = b.getShort();
            
            b.getShort();  //leafwaterID
            b.getInt(); //compress light cube
            b.getInt();
            b.getInt();
            b.getInt();
            b.getInt();
            b.getInt();
            
            b.getShort(); //padding?
            //System.out.println("Leaf "+i+" c "+leaf[i].contents+" cl "+leaf[i].cluster+" AF "+leaf[i].area_flags+
            //    " mins "+leaf[i].mins[0]+","+leaf[i].mins[1]+","+leaf[i].mins[2]+
            //    " maxs "+leaf[i].maxs[0]+","+leaf[i].maxs[1]+","+leaf[i].maxs[2]+
            //    " flf "+leaf[i].fstleafface+" nlf "+leaf[i].numleafface+
            //    " flb "+leaf[i].fstleafbrush+" nlb "+leaf[i].numleafbrush);
            
        }
        System.out.println(leaves);
    }
    
    public void loadlfaces(ByteBuffer b)
    {
        //LEAF FACE 16
        
        b.position(lump[16].ofs);
        lfaces = lump[16].len/Lump.size(16);
        
        lface = new short[lfaces];
        
        for(int i=0;i<lfaces;i++)
        {
            lface[i] = b.getShort();
            
        }
        System.out.println(lfaces);
        
    }
    
    public void loadlbrushes(ByteBuffer b)
    {
        //LEAF BRUSH 17
        
        b.position(lump[17].ofs);
        lbrushes = lump[17].len/Lump.size(17);
        
        lbrush = new short[lbrushes];
        
        for(int i=0;i<lbrushes;i++)
        {
            lbrush[i] = b.getShort();
            
        }
        System.out.println(lbrushes);
    }
    
    
    public void transmodels()
    {
        
        // Mark brush model edges & transform vertexes due to origin
        
        int iface, isedge, iedge;
        
        for(int i=1;i<entities;i++)		//skip 0th ent (worldspawn)
        {
            int imodel = ent[i].model;
            if(imodel<=0)
                continue; 	//not a brush model
            mod[imodel].entity = i;	// register entity with model
            
            
        }
    /* disable
            float ofx = ent[i].ox;
            float ofy = ent[i].oy;
            float ofz = ent[i].oz;
            //System.out.println("Model "+imodel+": ("+modff[imodel]+" "+modnf[imodel]+")");
            for(int j=0;j<mod[imodel].numface;j++)
            {
                iface = mod[imodel].fstface+j;
                for(int k=0;k<face[iface].numedge;k++)
                {
                    isedge = face[iface].fstedge+k;
                    iedge = sefix(isedge);
     
                    eim[iedge]=3;
                    //adjust vertexes by origin
                    int v1 = vi1[iedge]; int v2 = vi2[iedge];
                    if(!o[v1])
                    {
                        x[v1] += ofx;
                        y[v1] += ofy;
                        z[v1] += ofz;
                        o[v1] = true;
                    }
                    if(!o[v2])
                    {
                        x[v2] += ofx;
                        y[v2] += ofy;
                        z[v2] += ofz;
                        o[v2] = true;
                    }
     
     
                }
            }
        }
     
        //Check all edges for move/unmove mismatch:
        int numis=0;
        for(int i=0;i<edges;i++)
        {
            if(o[vi1[i]]!=o[vi2[i]])
            {
                numis++;
                vi1[i]=vi2[i];
            }
        }
        if(numis>0) System.out.println(numis+" edges required fixup");
     */
    }
    
    
    
    public void markedges()
    {
        //MARK MODEL EDGES
        
        int iface, isedge, iedge;
        
        for(int i=0;i<models;i++)
        {
            //System.out.println("Model "+i+": ("+modff[i]+" "+modnf[i]+")");
            for(int j=0;j<mod[i].numface;j++)
            {
                iface = mod[i].fstface+j;
                //System.out.println(" Face "+iface+": ("+facefe[iface]+" "+facene[iface]+")");
                for(int k=0;k<face[iface].numedge;k++)
                {
                    isedge = face[iface].fstedge+k;
                    iedge = sefix(isedge);  // fixup sedge -> edge & -ve index
                    
                    //System.out.print(iedge+" ");
                    eim[iedge]=(byte)(i==0?2:1);
                }
                //System.out.println();
            }
        }
        
    }
    
    public String readline(ByteBuffer bb, int end)
    {
        StringBuffer linebuff = new StringBuffer();
        char c;
        while((c=(char)bb.get())!='\n' && bb.position()<end)
        {
            linebuff.append(c);
        }
        return linebuff.toString();
    }
    
    public String readstr(ByteBuffer bb, int len)		// read len bytes, returning string up to first null
    {
        StringBuffer strbuff = new StringBuffer();
        char c;
        int pos=0;
        while((c=(char)bb.get())!=0 && ++pos<len)
            strbuff.append(c);			// read until null terminator
        
        while(++pos<len)	bb.get();			// read rest of null bytes to len
        
        return strbuff.toString();
    }
    
    public Texture gettex(short index)
    {
        return gettex(index, new Vec(0,0,0), new DecimalFormat("0.###"));
    }
    
    public Texture gettex(short index, Vec origin, DecimalFormat fp)
    {
        Texture tex = new Texture();
        
        if(index<0)
            return tex;
        Texinfo ti = tx[index];
        
        
        tex.texinfo = ti;
        if(ti.texdata<0)
		return tex;
        tex.texdata = td[ti.texdata];
        
        tex.name = texname[tex.texdata.nameid];
        
        tex.envmap = texenv[tex.texdata.nameid];
        
        float [][] tvec = ti.tv;
        double uaxis[]= new double[3];
        double vaxis[]= new double[3];
        double shift0, shift1, tw0, tw1, dotp0, dotp1;
        
        tw0 = 1.0 / Math.sqrt(Math.pow(tvec[0][0],2) + Math.pow(tvec[0][1],2) + Math.pow(tvec[0][2],2) );
        tw1 = 1.0 / Math.sqrt(Math.pow(tvec[1][0],2) + Math.pow(tvec[1][1],2) + Math.pow(tvec[1][2],2) );
        
        for(int i=0;i<3;i++)
        {
            uaxis[i] = tvec[0][i] * tw0;
            vaxis[i] = tvec[1][i] * tw1;
        }
        
        Vec t0 = new Vec(tvec[0][0], tvec[0][1], tvec[0][2]);
        Vec t1 = new Vec(tvec[1][0], tvec[1][1], tvec[1][2]);
        
        dotp0 = origin.dot(t0);
        dotp1 = origin.dot(t1);
        
        shift0 = tvec[0][3] - dotp0;
        shift1 = tvec[1][3] - dotp1;
        
        tex.uaxis = "["+fp.format(uaxis[0])+" "+fp.format(uaxis[1])+" "+fp.format(uaxis[2])+" "+fp.format(shift0)+"] "+fp.format(tw0);
        tex.vaxis = "["+fp.format(vaxis[0])+" "+fp.format(vaxis[1])+" "+fp.format(vaxis[2])+" "+fp.format(shift1)+"] "+fp.format(tw1);
        
        return tex;
    }
    
    public Wind windbrushface(Brush brush, int side)     // return a winding for brush side
    {                                                       // clipped to other sides of the brush
        Wind bw = new Wind();
        int bside = brush.fstside+side;
        int plane = bsd[bside].pnum;
        
        bw.createplane(pl[plane]);
        
        for(int i=0;i<brush.numside;i++) // clip to all other planes
        {
            if (i==side) continue;     // don't clip plane to itself
            int ibside = brush.fstside+i;
            int iplane = bsd[ibside].pnum;
            if(bsd[ibside].bevel == 1) continue;     // don't clip to bevel planes
            
            bw = bw.clipplane(pl[iplane].flip());   // clip to flipped plane
            
        }
        return bw;          // return the clipped winding
    }
    
    public void treewalk(int inode, Treelimit tl)
    {
        if(inode<0) // this is really a leaf
        {
            int ileaf = -1-inode;	// leaf index
            
            Leaf l = leaf[ileaf];
            for(int i=0;i<l.numleafface;i++)
            {
                int iface = lface[l.fstleafface+i];
                if(iface>tl.fmax)	tl.fmax=iface;
                if(iface<tl.fmin)	tl.fmin=iface;
            }
            for(int i=0;i<l.numleafbrush;i++)
            {
                int ibrush = lbrush[l.fstleafbrush+i];
                if(ibrush>tl.bmax)	tl.bmax=ibrush;
                if(ibrush<tl.bmin)	tl.bmin=ibrush;
            }
        }
        else		// this is a node
        {
            if(inode>tl.nmax)	tl.nmax = inode;
            
            treewalk(node[inode].child[0], tl);
            treewalk(node[inode].child[1], tl);
        }
    }
    
    
}

class Texture
{
    String name = "";
    Texinfo texinfo;
    Texdata texdata;
    int envmap = -1;
    String uaxis = "[1 0 0 0] 0.25";
    String vaxis = "[0 1 0 0] 0.25";
}


class Face
{
    int     pnum;
    byte    side;
    byte    onnode;
    int     fstedge;
    short   numedge;
    short   texinfo;
    short   dispinfo;
    int     orig;
    int     smooth;
    float   area;
    boolean written=false;      // for ofaces, if written to vmf already
    boolean undersize=false;    // for ofaces, if smaller than sum of faces
    boolean mark=false;         // for ofaces, for unique face counting
}

class	Model
{
    int	fstface;
    int	numface;
    int headnode;       // the head node of the model's BSP tree
    int entity; 	// index of entity which uses this model
    int fstbrush=-1;
    int numbrush;
}

class	Entity
{
    String classname;
    String targetname;
    String parentname;
    int model;
    float ox;
    float oy;
    float oz;
    ArrayList<String> key = new ArrayList<String>();
    ArrayList<String> value = new ArrayList<String>();
    ArrayList<Entity> link = new ArrayList<Entity>();	// link to other entity in this key/value
    
    int con = -1;	// if set, is the 1st kv pair that is a connection
    boolean mark;	// computation only marker
    int index;	// this entity's position in the array
    
    public String toString()		// display method: "classname (targetname)"
    {
        return ( this.classname+(this.targetname==null?"":(" ("+this.targetname+")")) );
    }
    
}

class Node
{
    int pnum;
    int child[] = new int[2];
    short mins[] = new short[3];
    short maxs[] = new short[3];
    short fstface;
    short numface;
    short area;
}

class Leaf
{
    int contents;
    short cluster;
    short area_flags;
    short mins[] = new short[3];
    short maxs[] = new short[3];
    
    short fstleafface;
    short numleafface;
    
    short fstleafbrush;
    short numleafbrush;
    
}

class Texinfo
{
    float tv[][] = new float[2][4];
    int	flags;
    int	texdata;
}

class Texdata
{
    int	nameid;
}

class Pstatic
{
    float ox, oy, oz;
    float ax, ay, az;
    int skin;
    short type;
    byte solid, flags;
    float fademin, fademax;
}

class GLump
{
    int id;
    short flags;
    short vers;
    int ofs;
    int len;
}

class Cubemap
{
    int x,y,z;
    byte size;
    HashSet<Integer> sides = new HashSet<Integer>();
}

class Plane
{
    float nx, ny, nz;
    float dist;
    int type;
    
    public Plane(Vec v, float d)        // helper functions
    {
        nx = v.x;   ny = v.y;   nz = v.z;
        dist = d;
    }
    
    public Plane()
    {
        this(new Vec(0,0,0),0);
    }
    
    public Plane flip()
    {
        Plane ret = new Plane(new Vec(-nx,-ny,-nz),-dist);
        return ret;
    }
    
    public Vec getvec()
    {
        return new Vec(nx, ny, nz);
    }
    
    public String toString()
    {
        return "("+nx+", "+ny+", "+nz+") ["+dist+"]";
    }
    
}

class Brush
{
    int fstside;
    int numside;
    int contents;
    
    public String toString()
    {
        return fstside+" "+numside+" "+contents;
    }
}

class Brushside
{
    int pnum;       // is ushort
    short texinfo;
    short dispinfo;
    short bevel;
}

class Dispinfo
{
    float sx, sy, sz;       // start position
    int ivert;              // index into disp verts
    int itri;               // index into disp tris
    int power;              // power (size)
    int mintess;            // min tesselation
    float smoothangle;      // smoothingangle
    int contents;           // surf contents
    short   face;           // map face
    int allow [] = new int[10];    // allowed verts
    
    public int psize()
    {
        return 1<<power;
    }
    
    public int numverts()
    {
        return (psize()+1)*(psize()+1);
    }
    
    public int numtris()
    {
        return 2*psize()*psize();
    }
}

class Dispvert
{
    float x, y, z;
    float dist;
    float alpha;
}

// Tris BSP=VMF 6=9  2=1  0=0

class Treelimit
{
    static final int HUGE = 2147483647;		// Maximum +int
    
    int bmin = HUGE;    //min brush
    int bmax = -1;      //max brush
    int fmin = HUGE;    //min face
    int fmax = -1;      //max face
    int nmax = -1;      //max node

    
    public void clear()
    {
        bmin = HUGE;
        bmax = -1;
        fmin = HUGE;
        fmax = -1;
        nmax = -1;
    }
}
