/*
 * Vec.java
 *
 * Created on January 20, 2005, 7:41 PM
 */

package unmap;

/**
 * A simple 3-vector class, stored at float accuracy
 * @author Bob
 */
public class Vec			// simple vector class
{
    float x, y, z;
    
    /**
     * Default constructor, equivalent to Vec(0.0, 0.0, 0.0)
     */
    public Vec()		// null constructor
    {
        this(0,0,0);
    }
    
    /**
     * Replicating contructor
     * Constructs new Vec by copying an existing Vec
     * @param v The Vec to copy
     */
    public Vec(Vec v)	// replicator constructor
    {
        this(v.x,v.y,v.z);
    }
    
    /**
     * Constructs a new Vec from x,y and z components
     * @param ox the vector x component
     * @param oy the vector y component
     * @param oz the vector z component
     */
    public Vec(float ox, float oy, float oz)	// standard constructor
    {
        x=ox;
        y=oy;
        z=oz;
    }
    
    /**
     * Vector dot product
     * @param that the Vec to take dot product with
     * @return the dot product of the two vectors
     */
    public float dot(Vec that)		// return dot product  this . that
    {
        return this.x * that.x + this.y * that.y + this.z * that.z;
    }
    
    /**
     * Vector cross product
     * @param that the vector to take a cross product
     * @return the cross-product vector
     */
    public Vec cross(Vec that)		// return cross product   this x that
    {
        Vec ret = new Vec();	// return vector
        
        ret.x = this.y * that.z - this.z * that.y;
        ret.y = this.z * that.x - this.x * that.z;
        ret.z = this.x * that.y - this.y * that.x;
        
        return ret;
    }
    
    /**
     * Vector normalisation
     * @return the normalised vector
     */
    public Vec norm()	// return normalized vector;	^this
    {
        float len;
        Vec ret = new Vec();
        
        len = modulus();
        
        ret.x = x/len;	ret.y = y/len;	ret.z  = z/len;
        return ret;
    }
    
    /**
     * Vector addition
     * @param that The vector to add
     * @return The sum of the two vectors
     */
    public Vec add(Vec that)		// return vector addition  this + that
    {
        Vec ret = new Vec();
        
        ret.x = x + that.x;
        ret.y = y + that.y;
        ret.z = z + that.z;
        return ret;
    }
    
    public Vec subtract(Vec that)		// return vector subtraction  this - that
    {
        Vec ret = new Vec();
        
        ret.x = x - that.x;
        ret.y = y - that.y;
        ret.z = z - that.z;
        return ret;
    }
    
    public void snap(float value)			// snap vector to nearest 'value'
    {
        x = Math.round(x/value) * value;
        y = Math.round(y/value) * value;
        z = Math.round(z/value) * value;
    }
    
    public boolean notvalid()           // check for any vec component being NaN
    {
        return Float.isNaN(x) || Float.isNaN(y) || Float.isNaN(z);
    }
    
    public float modulus()			// return modulus (length) of this vector
    {
        return (float)Math.sqrt(Math.pow(x,2) + Math.pow(y,2) + Math.pow(z,2));
    }
    
    public Vec scalar(float mul)        // return scalar multiplied vector
    {
        Vec ret = new Vec();
        ret.x = x*mul;
        ret.y = y*mul;
        ret.z = z*mul;
        
        return ret;
    }
    
    public float comp(int axis)     // return vector component along axis (x/y/z)
    {
	  switch(axis)
        {
    		case 0:
			return x;
		case 1:
			return y;
		case 2:
			return z;
	  }
	  return 0;			
    }

    public void setcomp(int axis, float value)  // sets vector component along axis (x/y/z)
    {
	  switch(axis)
        {
    		case 0:
			x = value;
                        break;
		case 1:
			y = value;
                        break;
		case 2:
			z = value;
                        break;
	  }
    }

    public Vec addmult(float scale, Vec that)       // returns this + scale*that
    {
	   Vec ret = new Vec();
	   ret.x = x + scale * that.x;
	   ret.y = y + scale * that.y;
	   ret.z = z + scale * that.z;

	   return ret;
    }
    
    public Vec minbound(Vec mins)
    {
        Vec ret = new Vec(mins);
        if(x < mins.x)  ret.x = x;
        if(y < mins.y)  ret.y = y;
        if(z < mins.z)  ret.z = z;
        return ret;
    }
    
    public Vec maxbound(Vec maxs)
    {
        Vec ret = new Vec(maxs);
        if(x > maxs.x)  ret.x = x;
        if(y > maxs.y)  ret.y = y;
        if(z > maxs.z)  ret.z = z;
        return ret;
    }

    public String toString()                    // stringifier
    {
        return "("+x+","+y+","+z+")";
    }

    public Vec copy()                           // returns clone of vector
    {
	  Vec ret = new Vec();
	  ret.x = x;
	  ret.y = y;
	  ret.z = z;
	
	  return ret;
    }
}