/*
 *Not Wind like the air, but wind like a watch
 *
 */

package unmap;

import java.util.*;
import java.text.*;

public class Wind
{
    ArrayList<Vec> v;			// list of vectors to vertex points
    
    static final int Maxlen = 56756;  // sqrt(3)*32768
    static final int Max_coord = 32768;
    static final int Side_front = 0;
    static final int Side_back = 1;
    static final int Side_on = 2;
    static final float Eps_split = 0.01f;
    static final float Eps_comp = 0.5f;
    
    public Wind()
    {
        v = new ArrayList<Vec>();
    }
    
    public int size()
    {
        return v.size();
    }
    
    public void createplane(Plane pl)		// creates a huge square winding from a Plane
    {
        // find the dominant axis of plane normal
        float dmax = -1.0f;
        int idir = -1;
        Vec norm = new Vec(pl.nx, pl.ny, pl.nz);
        
        for(int i=0;i<3;i++)			// for each axis
        {
            float dc = Math.abs(norm.comp(i));
            if( dc > dmax )			// find the biggest component
            {
                dmax = dc;
                idir = i;
            }
        }
        
        if(idir == -1)				// didn't find one (null or NaN'ed vector)
        {
            System.out.println("createplane: bad plane normal");
            return;
        }
        
        Vec vup = new Vec(0,0,0);		// this will be the "upwards" pointing vector
        
        switch(idir)
        {
            case 0:				// if x or y dominant
            case 1:
                vup.z = 1.0f;		// use z unit vector
                break;
            case 2:				// if z dominant
                vup.x = 1.0f;		// use x unit vector
                break;
        }
        
        float vdot = vup.dot(norm);
        
        vup = vup.addmult(-vdot,norm);
        vup = vup.norm();
        
        Vec org = norm.scalar(pl.dist);	// the vector from origin perpendicularly touching plane
        
        Vec vrt = vup.cross(norm);		// this is the "rightwards" pointing vector
        
        vup = vup.scalar(Maxlen);
        vrt = vrt.scalar(Maxlen);
        
        v.add( org.subtract(vrt).add(vup) );
        
        v.add( org.add(vrt).add(vup) );
        
        v.add( org.add(vrt).subtract(vup) );
        
        v.add( org.subtract(vrt).subtract(vup) );
        
        
    }
    
    public Wind clipplane(Plane pl)	// clips a winding to a plane, returning the clipped winding
    {						// use after createplane()
        //System.out.println("XP:"+pl);
        Vec norm = new Vec(pl.nx, pl.ny, pl.nz);	// the clip plane normal
        int counts[] = new int[3];	// counts number of Front, Back and On vetices
        Wind w = new Wind();		// the clipped return winding
        
        float dists[] = new float[this.size()+1];
        int sides[] = new int[this.size()+1];
        
        counts[0] = counts[1] = counts[2] = 0;
        
        for(int i=0; i < this.size(); i++)
        {
            float dot = v.get(i).dot(norm);	// distance along norm-dirn from origin to vertex
            dot -= pl.dist;				// distance along norm-dirn from clip plane to vertex
            dists[i] = dot;				// store it
            if( dot > Eps_split)			// vertex in front of plane
                sides[i] = Side_front;
            else if (dot < -Eps_split)		// vertex behind plane
                sides[i] = Side_back;
            else
                sides[i] = Side_on;		// vertex on plane (within epsilon)
            
            //System.out.println("XC:"+i+" "+v.get(i)+" "+sides[i]);
            
            counts[sides[i]]++;			// count relative vertex positions
        }
        
        sides[this.size()] = sides[0];		// loop around to 0'th
        dists[this.size()] = dists[0];
        
        if(counts[Side_front]==0 && counts[Side_back]==0)	// no front or back vertices - winding is planar to clip
        {
            //System.out.println("Planar");
            return this;						// return unchanged
        }
        
        if(counts[Side_front]==0)	// no vertices in front - all behind clip plane
        {
            //System.out.println("Empty");
            return w;			// return empty winding
        }
        if(counts[Side_back]==0)	// no vertices in back - all in front of clip plane
        {
            //  System.out.println("Full");
            return this;		// return unchanged
        }
        
        for(int i=0; i< this.size(); i++)
        {
            Vec p1 = v.get(i);	// get i'th vertex
            int mi = w.size();	// index to next place to add vertex in output winding
            Vec mv = new Vec(0,0,0);// the vertex to add
            
            if(sides[i] == Side_front || sides[i] == Side_on)  // this vertex in front of or on clip plane
            {
                mv = p1.copy();	// copy the current vertex
                //System.out.print("A");
                w.vset(mi,mv);	// write it to ouput winding at mi'th position
                if(sides[i] == Side_on)	// this vertex is on the clip plane
                    continue;		// so don't inc out pointer?
                mi = w.size();		// inc outpointer to last entry
                mv = new Vec(0,0,0);	// need to check this
            }
            
            if(sides[i+1] == Side_on || sides[i+1] == sides[i])	// next vertex is on the plane, or does not change side
                continue;		// so go to next vertex stat
            // otherwise, we are crossing the clip plane between this vertex and the next
            // so generate a split point
            
            Vec p2;				// will contain the next vertex position
            int pi2;				// will contain index of next vertex
            
            if (i== this.size()-1)		// if we're the last vertex in the winding
            {
                p2 = v.get(0);		// next vertex is the 0'th one
                pi2 = 0;			// set index
            }
            else
            {
                p2 = v.get(i+1);		// else get the next vertex
                pi2 = i+1;			// set index
            }
            
            // mi++;					// inc output index
            
            float dot = dists[i] / (dists[i]-dists[i+1]);	// dot is fractional position of clip plane between
            // this vertex and the next
            for(int j=0; j<3; j++)		// for each axis
            {				// avoid rounding errors?
                if(norm.comp(j) == 1.0f)
                    mv.setcomp(j, pl.dist);
                else if(norm.comp(j) == -1.0f)
                    mv.setcomp(j, -pl.dist);
                else	// check it! MSH
                    mv.setcomp(j, p1.comp(j) + dot * (p2.comp(j) - p1.comp(j)));
            }
            //System.out.print("B");
            
            w.vset(mi, mv);	// write the output vertex
        }
        return w;
    }
    
    public void vset(int index, Vec vertex)
    {
        int size = v.size();
        
        //System.out.println("Vset: "+index+" "+size);
        if(size==index)
        {
            v.add(vertex);
        }
        else if (size>index)
        {
            v.set(index, vertex);
        }
        else
        {
            System.out.println("Vset: index "+index+" > size "+size);
        }
        //System.out.println("Vset: "+this);
    }
    
    public String toString()
    {
        StringBuffer s = new StringBuffer();
        
        s.append(size()+" : ");
        for(int i=0;i<size();i++)
            s.append(" ("+v.get(i).x+", "+v.get(i).y+", "+v.get(i).z+") ");
        
        return s.toString();
    }
    
    public boolean matches(Wind that)       // compare two windings, taking into account that start points may not match
    {
        if(this.size() != that.size())      // if windings have different number of points, trivially fail
            return false;                   //  note we don't yet crunch winding vertices, so this may give false negatives
        
        float min = 1e6f; // minimum match distance
        
        for(int i=0;i<this.size();i++)
        {
            float mdist = this.matchdist(that, i);   // get the aggregate distance at offset i
            if(mdist<min) min = mdist;
        }
        
        if(min<Eps_comp)        // match was close enough
            return true;
        else
            return false;
    }
    
    public float matchdist(Wind that, int offset)
    {
        float dist = 0;    
        for(int i=0;i<size();i++)
        {
            int j = (i+offset)%size();      // wrap index if greater than size
            dist += this.v.get(i).subtract(that.v.get(j)).modulus();        // distance between vertex i of this and j of that
        }
        return dist;
    }
    
    public float vmatches(Wind that)       // compare two windings, taking into account that start points may not match
    {                                         // returns the min distance aggregate
        
        float min = 1e6f; // minimum match distance
        
        if(this.size() != that.size())      // if windings have different number of points, trivially fail
            return min;                   //  note we don't yet crunch winding vertices, so this may give false negatives
        
        
        for(int i=0;i<this.size();i++)
        {
            float mdist = this.matchdist(that, i);   // get the aggregate distance at offset i
            if(mdist<min) min = mdist;
        }
        
        return min;
    }
    
    
    //factory method - construct a winding from a map face
    static public Wind windfromface(BSP m, Face f)
    {
        Wind w = new Wind();
        
        for(int i=0;i<f.numedge;i++)
        {
            int sedge = m.sedge[f.fstedge+i];
            int v = 0;
            
            if(sedge<0)
                v = m.vi2[-sedge];
            else
                v = m.vi1[sedge];
            
            w.v.add(new Vec(m.x[v], m.y[v], m.z[v]));
        }
        return w;
    }
    
    //factory method - construct a winding from a brush, for a brush side
    public static Wind windfromside(BSP m, Brush brush, int side)     // return a winding for brush side
    {                                                               // clipped to other sides of the brush
        Wind bw = new Wind();
        int bside = brush.fstside+side; 
        int plane = m.bsd[bside].pnum;
        
        bw.createplane(m.pl[plane]);
        
        for(int i=0;i<brush.numside;i++) // clip to all other planes
        {
            if (i==side) continue;     // don't clip plane to itself
            int ibside = brush.fstside+i;
            int iplane = m.bsd[ibside].pnum;
            if(m.bsd[ibside].bevel == 1) continue;     // don't clip to bevel planes
            
            bw = bw.clipplane(m.pl[iplane].flip());
            
        }
        return bw;          // return the clipped winding
    }
    
    public boolean isnull()
    {
        return (size()==0);
    }
    
    public boolean ishuge()
    {
        Iterator<Vec> vit = v.iterator();
        
        while(vit.hasNext())
        {
            Vec point = vit.next();
            for(int j=0;j<3;j++)
            {
                if(Math.abs(point.comp(j))>Max_coord)
                    return true;
            }
        }
        return false;
    }
          
}

