/*
 *
 * Vmex.java
 *
 * A simple decompiler for HL2 bsp files
 */

package unmap;

import java.lang.*;
import java.util.*;
import java.util.regex.*;
import java.text.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;

public class Vmex
{
    
    // parameters derived from arguments
    String filename;
    int	 dmode=3;		// decomile mode; 0=ofaces, 1=faces, 2=original unless missing, 3= brushes
    int	 bmode=0;		// brush mode; 0=pyrimidal, 1=prismatic, 2=face only
    float depth=1.0f;           // depth of flat
    int stexmode=2;		// back face texture;  0=white, 1=black, 2=invisible, 3=face tex
    int ttexmode=3;		// front face texture; 0=white, 1=black, 2=orangegen, 3=face tex
    //boolean entonly=false;	// output only ents
    //boolean snap=false;	// snap vertices
    //float	snapval=0.0f;	// snap to nearest
    int	prec=8;         	// decimal precision used for outputing floats
    boolean fixorg=true;	// fixup model origins in world
    boolean fixtex=true;	// fixup texture names of envmapped materials
    boolean props=true;         // output prop_statics
    boolean debug=false;        // debugging output
    boolean checkverts=true;	// check vertices for best normal
    boolean details=true;
    boolean disps=true;	// output displacements
    int envmaps=8;             // output cubemaps;  0=no, 1=location only, n list sides if <n are mapped
    
    int brushid=2;		// the brush id (1=worldspawn)
    int sideid=1;		// the side id
    
    BSP m;			// the map file
    PrintWriter p;		// the output printwriter
    
    DecimalFormat fp;		// the output decimal precision
    Vec org;			// the map origin (0,0,0)
    
    HashSet [] origmap;  // set of face indicies which correspond to ofaces
    float area_eps = 1.0f;     // epsilon for area comparison slop, in mu^2
    int oucount = 0;
    int worldbrushes;          // number of world brushes. This is index of first model brush
    
    static int PLAYERCLIP = 0x10000;            // Brush contents
    static int NPCCLIP = 0x20000;
    static int CLIP = PLAYERCLIP | NPCCLIP;
    static int DETAIL = 0x8000000;
    static int SOLID = 0x1;
    
    static String VERSION = "v0.91";
    
    public void exec()
    {
        try
        {
            long starttime = System.currentTimeMillis();	// System start time (in ms)
            
            File infile = new File(filename);
            if(!infile.exists() || !infile.canRead())
            {
                System.out.println("Can't read " +filename);
                return;
            }
            
            System.out.println("Reading "+filename);
            
            FileChannel raf = new RandomAccessFile(infile,"r").getChannel();
            MappedByteBuffer mbuffer = raf.map( FileChannel.MapMode.READ_ONLY, 0, raf.size() );
            ByteBuffer buff = mbuffer.order( ByteOrder.LITTLE_ENDIAN ) ;
            
            //Load the map
            
            m = new BSP();	// new instance of the BSP
            
            m.loadmap(buff);
            
            raf.close();	// close infile
            
            //Argument option setup
            
            String precfmt = "0.";
            for(int i=0;i<prec;i++)
                precfmt = precfmt + "#";	// add a # for each digit of precision
            
            Locale.setDefault(Locale.ENGLISH);
            fp = new DecimalFormat(precfmt);
            
            org = new Vec(0,0,0);		// map origin
            
            fixtexturenames();
            
            if(dmode==2)    buildfacemaps();
            
            protbrushcheck();
            
            // *** disable protection here
            if(m.no_decomp)
            {
                System.out.println("**** This map is protected from decompiling");
                System.exit(0);
            }
            
            
            //Write the decompiled map
            
            String fullname = infile.getPath();
            String outfullname;		// the outfile fqpn
            
            int index;
            if( (index=fullname.lastIndexOf(".bsp"))>0)	// build outfilename
            {
                outfullname = fullname.substring(0,index) + "_d.vmf";
            }
            else
                outfullname = fullname+"_d.vmf";
            
            File outfile = new File(outfullname);		// the file to write
            FileWriter ofw  = new FileWriter(outfile);	// output file writer
            p = new PrintWriter(ofw);		// output print writer
            
            System.out.println("\nWriting "+outfile.getName());
            
            
            System.out.print("Decompiling via ");
            if(dmode==0)	System.out.println("original faces");
            if(dmode==1)	System.out.println("split faces");
            if(dmode==2)	System.out.println("original faces where possible");
            if(dmode==3)        System.out.println("brushes and planes");
            
            if(dmode==3)
            {
                System.out.println("Assigning brushes to models");
                assignbrushes();
            }
            
            System.out.println("\nWorld ");
            vmfheader();
            vmfworldspawn();
            
            switch(dmode)
            {
                case 0:
                    vmforigfaces();
                    break;
                case 1:
                    vmffaces();
                    break;
                case 2:
                    vmforigfacesplus();
                    break;
                case 3:
                    vmfbrushes();
                    break;
            }
            if(disps)
            {
                System.out.println("\nDisplacements ");
                vmfdisplacements();
            }
            
            System.out.println("\nEntities ");
            vmfentities();
            
            if(details)
            {
                System.out.println("\nFunc_details ");
                vmfdetails();
            }
            
            
            if(props)
            {
                System.out.println("\nProp_statics ");
                vmfstatics();
            }
            
            if(envmaps>0)
            {
                System.out.println("\nEnv_cubmaps ");
                vmfcubemaps();
            }
            
            vmftrailer();
            
            if(dmode==2)    System.out.println("\n"+oucount+" original faces were written as split faces");
            
            ofw.close();			// close the outfile
            
            long duration = System.currentTimeMillis() - starttime;	// number of ms elapsed
            DecimalFormat tf = new DecimalFormat("0.#");
            
            System.out.println("\nDone in "+tf.format((float)(duration/1000.0f))+" seconds");
            
        }
        catch(Exception excession)
        {
            System.out.println(excession);
            return;
        }
    } // exec
    
    
    public void vmfdisplacements()
    {
        for(int i=0;i<m.ofaces;i++)
        {
            short disp = m.oface[i].dispinfo;   // the dispinfo index
            if(disp<0)  continue;               // if -1, skip
            
            Face of = m.oface[i];
            Dispinfo di = m.dinfo[disp];
            
            System.out.print(".");
            
            Vec org = new Vec();
            
            // Write the back faces, but first compute the front norm
            
            Vec [] svec = new Vec[of.numedge];	// list of vectors
            int [] vert = new int[of.numedge];	// list of vertices
            
            int sedge = m.sedge[of.fstedge];
            int v0;					// the start vertex of the face
            if(sedge<0)
                v0 = m.vi2[-sedge];		// backwards wound edge
            else
                v0 = m.vi1[sedge];		// forwards wound edge
            
            float x0 = m.x[v0];
            float y0 = m.y[v0];
            float z0 = m.z[v0];
            
            for(int j=0;j<of.numedge;j++)		// build vector list
            {
                sedge = m.sedge[of.fstedge+j];
                // j'th vertex in face
                if(sedge<0)
                    vert[j] = m.vi2[-sedge];
                else
                    vert[j] = m.vi1[sedge];
                
                svec[j] = new Vec(m.x[vert[j]]-x0, m.y[vert[j]]-y0, m.z[vert[j]]-z0);  // the vector from start vertex to j'th
            }
            
            float maxmcp = -1.0f;	// the largest modulus of cross product found between ixj
            int jmax = -1;	// the j index of largest cp
            int kmax = -1;	// the k index of largest cp
            
            for(int j=1;j<of.numedge;j++)					// loop through all i x j combinations
            {
                for(int k=j+1;k<of.numedge;k++)			// ensures k>j
                {
                    float mcp = svec[j].cross(svec[k]).modulus();	//
                    if(mcp>maxmcp)
                    {
                        maxmcp = mcp;
                        jmax = j;
                        kmax = k;
                    }
                }
            }
            
            int v1 = vert[0];	// 0'th vert is always base vertex
            int v2 = vert[jmax];	//
            int v3 = vert[kmax];	// choose other two such that cross product is maximum
            
            Vec e1, e2, e3;		// shift vertices by origin vector
            
            e1 = new Vec(m.x[v1],m.y[v1],m.z[v1]);
            e2 = new Vec(m.x[v2],m.y[v2],m.z[v2]);
            e3 = new Vec(m.x[v3],m.y[v3],m.z[v3]);
            // *** round vertex coordinates here
            
            Vec ev12 = e2.subtract(e1);
            Vec ev13 = e3.subtract(e1);
            Vec norm = ev12.cross(ev13).norm();
            
            if(norm.notvalid())     // bad plane normal
            {
                System.out.print("\nBad normal in displacement! "+ev12+"x"+ev13);
            }
            
            String sv1 = strpvec(e1);
            String sv2 = strpvec(e2);
            String sv3 = strpvec(e3);
            
            Texture tex = m.gettex(of.texinfo, org, fp);
            
            String texture = "NULL";
            if(ttexmode==0)	texture = "SKYBOX/SKYFAKEWHITE";
            if(ttexmode==1) texture = "TOOLS/TOOLSBLACK";
            if(ttexmode==2)	texture = "DEV/DEV_MEASUREGENERIC01";
            if(ttexmode==3)	texture = tex.name;
            
            String btexture = "NULL";
            if(stexmode==0)	btexture = "SKYBOX/SKYFAKEWHITE";
            if(stexmode==1)     btexture = "TOOLS/TOOLSBLACK";
            if(stexmode==2)	btexture = "TOOLS/TOOLSINVISIBLE";
            if(stexmode==3)	btexture = texture;
            
            String sua = tex.uaxis;
            String sva = tex.vaxis;
            
            tab(1); p.println("solid");
            tab(1); p.println("{");
            
            pl(2,"id",String.valueOf(brushid++));	// brush id
            
            writepyramback(of, 1, norm, org, btexture, sua, sva);
            
            // WRITE THE FRONT FACE
            
            tab(2);	p.println("side");
            tab(2);	p.println("{");
            
            pl(3, "id", String.valueOf(sideid++));	// side id
            pl(3, "plane", sv1+" "+sv2+" "+sv3);
            pl(3, "material", texture);
            pl(3, "uaxis", sua);
            pl(3, "vaxis", sva);
            pl(3, "rotation","0");
            pl(3, "lightmapscale", "16");
            pl(3, "smoothing_groups", String.valueOf(of.smooth));
            
            // dispinfo
            
            tab(3); p.println("dispinfo");
            tab(3); p.println("{");
            
            pl(4, "power", String.valueOf(di.power));
            pl(4, "startposition", "["+strvec(new Vec(di.sx,di.sy,di.sz))+"]");
            pl(4, "elevation", "0");    // ?
            pl(4, "subdiv", "0");       // probably wrong
            
            tab(4); p.println("normals");
            tab(4); p.println("{");
            
            Dispvert dv;
            int row=0;
            StringBuffer t = new StringBuffer();
            for(int j=0;j<di.numverts();j++)
            {
                dv = m.dv[di.ivert+j];
                
                t.append(fp.format(dv.x)+" "+fp.format(dv.y)+" "+fp.format(dv.z)+" ");
                
                if(j%(di.psize()+1)==di.psize())
                {
                    pl(5,"row"+String.valueOf(row++), t.toString() );
                    t = new StringBuffer();
                }
            }
            tab(4); p.println("}");
            
            tab(4); p.println("distances");
            tab(4); p.println("{");
            row=0;
            t = new StringBuffer();
            for(int j=0;j<di.numverts();j++)
            {
                dv = m.dv[di.ivert+j];
                
                t.append(fp.format(dv.dist)+" ");
                
                if(j%(di.psize()+1)==di.psize())
                {
                    pl(5,"row"+String.valueOf(row++), t.toString() );
                    t = new StringBuffer();
                }
            }
            tab(4); p.println("}");
            
            String onorm = "0 0 0";         // strvec(m.pl[of.pnum].getvec());
            
            tab(4); p.println("offset_normals");
            tab(4); p.println("{");
            row=0;
            t = new StringBuffer();
            for(int j=0;j<di.numverts();j++)
            {
                dv = m.dv[di.ivert+j];
                
                t.append(onorm+" ");
                
                if(j%(di.psize()+1)==di.psize())
                {
                    pl(5,"row"+String.valueOf(row++), t.toString() );
                    t = new StringBuffer();
                }
            }
            
            tab(4); p.println("}");
            
            tab(4); p.println("alphas");
            tab(4); p.println("{");
            row=0;
            t = new StringBuffer();
            for(int j=0;j<di.numverts();j++)
            {
                dv = m.dv[di.ivert+j];
                
                t.append(fp.format(dv.alpha)+" ");
                
                if(j%(di.psize()+1)==di.psize())
                {
                    pl(5,"row"+String.valueOf(row++), t.toString() );
                    t = new StringBuffer();
                }
            }
            
            tab(4); p.println("}");
            //BSP=VMF 6=9  2=1  0=0
            String [] triconvert = { "0", "0", "1", "0", "0", "0", "9" };
            
            tab(4); p.println("triangle_tags");
            tab(4); p.println("{");
            row=0;
            t = new StringBuffer("");
            for(int j=0;j<di.numtris();j++)
            {
                short dt = m.dtri[di.itri+j];
                
                t.append(triconvert[dt]+" ");
                
                if(j%(2*di.psize())==(2*di.psize()-1))
                {
                    pl(5,"row"+String.valueOf(row++), t.toString() );
                    t = new StringBuffer();
                }
            }
            
            tab(4); p.println("}");
            
            tab(4); p.println("allowed_verts");
            tab(4); p.println("{");
            
            tab(5); p.print("\"10\" \"");
            for(int j=0;j<10;j++)
                p.print(di.allow[j]+" ");
            p.println("\"");
            
            tab(4); p.println("}");
            
            tab(3); p.println("}");
            
            tab(2); p.println("}");
            
            tab(1); p.println("}");
            
        }
        p.println("}");
    }
    
    public void assignbrushes()        // assign brushes to models
    {                                     // calculate last world brush
        // walk the BSP tree
        // to from the headnode of each model
        // to calculate the minimum and maximum brush in the tree
        // much simpler than the guessing method
        // plus this recovers null-faced brushes
        
        Treelimit tl = new Treelimit();     // structure to record min/max of faces, brushes and nodes in the tree
        
        m.treewalk(0,tl);       // walk model0 (worldspawn model)
        
        worldbrushes = tl.bmax;
        
        for(int i=1;i<m.models;i++)
        {
            Model model = m.mod[i];
            tl.clear();
            m.treewalk(model.headnode, tl);
            
            model.fstbrush = tl.bmin;
            model.numbrush = tl.bmax - tl.bmin + 1;
            if(debug)   System.out.println("Model "+i+": Brush "+tl.bmin+" to "+tl.bmax+"\n");
            
        }
        System.out.println("\nLargest worldbrush: "+worldbrushes);
        
    }
    
    
    public void vmfbrushes()        // output map by brushes and planes
    {
        for(int i=0;i<worldbrushes;i++)     // all brushes up to last world brush
        {
            if(debug) System.out.print("\nBrush:"+i);
            
            if(!(m.br[i].contents == (SOLID|DETAIL)) || !details)       // isnt a detail brush or not using func_detail?
            {
                writebrush(1, m.br[i], org);    // write the brush
            }
            
        }
        if(!disps) p.println("}");	// end of world brushes (only if no disps)
    }
    
    
    public void vmfdetails()        // output func_details
    {
        for(int i=0;i<worldbrushes;i++)     // all brushes up to last world brush
        {
            
            if(m.br[i].contents == (SOLID|DETAIL))       // is a detail brush?
            {
                if(debug) System.out.print("\nFDBrush:"+i);
                
                p.println("entity");
                p.println("{");
                
                pl(1,"id",String.valueOf(brushid++));
                pl(1,"classname","func_detail");
                writebrush(1, m.br[i], org);    // write the brush
                
                p.println("}");
                
            }
            
            
        }
        
    }
    
    public String contentstr(int v)
    {
        String [] cont = {"SOLID","WINDOW","AUX","GRATE","SLIME","WATER","MIST","OPAQUE","TESTFOG",
                "U3","U4","U5","U6","U7","MOVABLE","AREAPORTAL","PLAYERCLIP","MONSTERCLIP",
                "CUR0","CUR90","CUR180","CUR270","CURUP","CURDN",
                "ORIGIN","MONSTER","DEBRIS","DETAIL","TRANSLUCENT","LADDER","HITBOX" };
                StringBuffer s = new StringBuffer();
                for(int i=0;i<32;i++)
                {
                    if((v & 0x1)==1) s.append(cont[i]+" ");
                    v = v >>> 1;
                }
                return s.toString();
    }
    
    public void writebrush(int td, Brush b, Vec o)
    {
        int visgrp = 0;
        tab(td); p.println("solid");
        tab(td); p.println("{");
        
        pl(td+1,"id",String.valueOf(brushid++));	// brush id
        
        if(debug) System.out.print(" "+contentstr(b.contents)+" "+b.fstside+" ("+b.numside+")");
        
        int nullsides = 0;
        
        for(int i=0;i<b.numside;i++)
        {
            Brushside bside = m.bsd[b.fstside+i];
            if(bside.bevel==1) continue;
            Wind w = Wind.windfromside(m, b, i);
            
            if(w.ishuge())  System.out.println("Side "+i+" of brush "+b+" is huge");
            
            if(w.isnull())
                nullsides++;
            else
            {
                int surf = writeside(2, m.bsd[b.fstside+i], w, o, b.contents);
                if(surf>0) visgrp = surf;
            }
            
        }
        if(nullsides == b.numside)
            System.out.println("Brush "+b+" is null");
        
        if(visgrp>0)    // was a clip or hint brush
        {               // place in correct visgroup
            tab(2); p.println("editor");
            tab(2); p.println("{");
            pl(3, "color", "200 100 100");
            pl(3, "visgroupid", String.valueOf(5+visgrp));
            pl(3, "visgroupshown", "1");
            tab(2); p.println("}");
        }
        
        tab(td); p.println("}");        // end of brush
        
    }
    
    
    public int writeside(int td, Brushside bs, Wind w, Vec origin, int cont)
    {
        int surface = 0;
        // Plane pn = m.pl[bs.pnum];
        if(bs.bevel ==1)
        {
            System.out.print("!");
            
            return 0;     // don't output bevel faces - they lead to bad brushes
        }
        
        System.out.print(".");
        
        int verts = w.size();
        
        Vec [] svec = new Vec[verts];	// list of vectors
        int [] vert = new int[verts];	// list of vertices
        
        Vec v0 = w.v.get(0);        // base vertex location
        
        for(int i=0;i<verts;i++)		// build vector list
        {
            svec[i] = w.v.get(i).subtract(v0);  // the vector from start vertex to i'th
        }
        
        float maxmcp = -1.0f;	// the largest modulus of cross product found between ixj
        int imax = -1;	// the i index of largest cp
        int jmax = -1;	// the j index of largest cp
        
        for(int i=1;i<verts;i++)			// loop through all i x j combinations
        {
            for(int j=i+1;j<verts;j++)		// ensures j>i
            {
                float mcp = svec[i].cross(svec[j]).modulus();	//
                if(mcp>maxmcp)
                {
                    maxmcp = mcp;
                    imax = i;
                    jmax = j;
                }
            }
        }
        
        Vec e1 = v0;	// 0'th vert is always base vertex
        Vec e2 = w.v.get(imax); 	//
        Vec e3 = w.v.get(jmax);         // choose other two such that cross product is maximum
        
        e1 = e1.add(origin);        // shift due to origin
        e2 = e2.add(origin);
        e3 = e3.add(origin);
        
        if(e1.notvalid() || e2.notvalid() || e3.notvalid())
        {
            System.out.println("\nBrushside with wind "+w+" is invalid");
            
        }
        
        String sv1 = strpvec(e1);
        String sv2 = strpvec(e2);
        String sv3 = strpvec(e3);
        
        String sua, sva, texture;
        
        Texture tex = m.gettex(bs.texinfo, origin, fp);
        
        if(tex.texinfo==null)   // not a valid texinfo
        {                   // use skip texture by default
            Vec n = m.pl[bs.pnum].getvec();     // plane normal
                       
            Vec udir = new Vec(0,0,1.0f);              // z is z-direction unit vector
            
            if(Math.abs(udir.dot(n))>0.8)        // if angle of normal to z axis is less than ~25 degrees
            {
                udir = new Vec(1.0f,0,0);  // use x unit-vector as basis
            }   // else use z unit-vector as basis
            
            Vec tv1 = udir.cross(n).norm();         // 1st tex vector
            Vec tv2 = tv1.cross(n).norm();            // 2nd tex vector
                       
            sua = "["+strvec(tv1)+" 0] 0.25";
            sva = "["+strvec(tv2)+" 0] 0.25";
            texture = "TOOLS/TOOLSSKIP";        // SKIP by default
            
            
            if((cont & CLIP) == PLAYERCLIP)    texture = "TOOLS/TOOLSPLAYERCLIP";
            if((cont & CLIP) == NPCCLIP)       texture = "TOOLS/TOOLSNPCCLIP";
            if((cont & CLIP) == CLIP)          texture = "TOOLS/TOOLSCLIP";
            surface = 1;
        }
        else
        {
            texture = "NULL";
            if(ttexmode==0)	texture = "SKYBOX/SKYFAKEWHITE";
            if(ttexmode==1) texture = "TOOLS/TOOLSBLACK";
            if(ttexmode==2)	texture = "DEV/DEV_MEASUREGENERIC01";
            if(ttexmode==3)	texture = tex.name;
            
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSHINT"))    surface = 2;
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSPLAYERCLIP"))  surface = 1;
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSNPCRCLIP"))  surface = 1;
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSCLIP"))  surface = 1;
            
            //check if texture used envmap
            
            if(envmaps>1)       // if outputting cubemaps with side lists
            {
                int envmapid = tex.envmap;
                if(envmapid!=-1) // if a cubemap is used for this texture
                    m.cm[envmapid].sides.add(new Integer(sideid));         // add the sideid to the cubemap list
                
            }
            
            sua = tex.uaxis;
            sva = tex.vaxis;
        }
        
        tab(td);    p.println("side");
        tab(td);    p.println("{");
        
        pl(td+1, "id", String.valueOf(sideid++));       // side id
        pl(td+1, "plane", sv1+" "+sv2+" "+sv3);
        pl(td+1, "material", texture);
        pl(td+1, "uaxis", sua);
        pl(td+1, "vaxis", sva);
        pl(td+1, "rotation","0");
        pl(td+1, "lightmapscale", "16");
        pl(td+1, "smoothing_groups", "0");
        
        tab(td);	p.println("}");
        return surface;
        
    }
    
    
    // not used
    public int writeside(int td, Brushside bs, Vec origin, int cont)
    {
        int surface = 0;
        Plane pn = m.pl[bs.pnum];
        if(bs.bevel ==1)
        {
            System.out.print("!");
            return 0;     // don't output bevel faces - they lead to bad brushes
        }
        
        System.out.print(".");
        Vec n = new Vec(pn.nx,pn.ny,pn.nz);     // plane normal
        float d = pn.dist;                      // distance to origin
        
        Vec e1 = n.scalar(d);                    // e1 is now vector from origin to plane, normal to the surface
        
        Vec udir = new Vec(0,0,1.0f);              // z is z-direction unit vector
        
        if(Math.abs(udir.dot(n))>0.8)        // if angle of normal to z axis is less than ~25 degrees
        {
            udir = new Vec(1.0f,0,0);  // use x unit-vector as basis
        }   // else use z unit-vector as basis
        
        Vec v1 = udir.cross(n).norm();         // 1st p vector
        // must be perp to normal, so is in plane
        Vec v2 = v1.cross(n).norm();            // 2nd p vector
        // perp to normal and v1, so is also in plane
        v1 = v1.scalar(100.0f);
        v2 = v2.scalar(100.0f);
        
        Vec e2 = e1.add(v1);        // plane vertices
        Vec e3 = e1.add(v2);
        
        e1 = e1.add(origin);        // shift due to origin
        e2 = e2.add(origin);
        e3 = e3.add(origin);
        
        if(e1.notvalid() || e2.notvalid() || e3.notvalid())
        {
            System.out.print("\n"+n+" "+udir+" "+udir.dot(n)+" "+e1+" "+d+" "+v1+" "+v2);
        }
        
        String sv1 = strpvec(e1);
        String sv2 = strpvec(e2);
        String sv3 = strpvec(e3);
        
        String sua, sva, texture;
        
        Texture tex = m.gettex(bs.texinfo, origin, fp);
        
        if(tex.texinfo==null)   // not a valid texinfo
        {                   // use skip texture by default
            Vec tv1 = udir.cross(n).norm();     // align it normal to plane, upwards or x-wards if near-horizontal
            Vec tv2 = v1.cross(n).norm();
            sua = "["+strvec(tv1)+" 0] 0.25";
            sva = "["+strvec(tv2)+" 0] 0.25";
            texture = "TOOLS/TOOLSSKIP";        // SKIP by default
            
            if((cont & CLIP) == PLAYERCLIP)    texture = "TOOLS/TOOLSPLAYERCLIP";
            if((cont & CLIP) == NPCCLIP)       texture = "TOOLS/TOOLSNPCCLIP";
            if((cont & CLIP) == CLIP)          texture = "TOOLS/TOOLSCLIP";
            surface = 1;
        }
        else
        {
            texture = "NULL";
            if(ttexmode==0)	texture = "SKYBOX/SKYFAKEWHITE";
            if(ttexmode==1) texture = "TOOLS/TOOLSBLACK";
            if(ttexmode==2)	texture = "DEV/DEV_MEASUREGENERIC01";
            if(ttexmode==3)	texture = tex.name;
            
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSHINT"))    surface = 2;
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSPLAYERCLIP"))  surface = 1;
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSNPCRCLIP"))  surface = 1;
            if(tex.name.equalsIgnoreCase("TOOLS/TOOLSCLIP"))  surface = 1;
            
            //check if texture used envmap
            
            if(envmaps>1)       // if outputting cubemaps with side lists
            {
                int envmapid = tex.envmap;
                if(envmapid!=-1) // if a cubemap is used for this texture
                    m.cm[envmapid].sides.add(new Integer(sideid));         // add the sideid to the cubemap list
                
            }
            
            sua = tex.uaxis;
            sva = tex.vaxis;
        }
        
        tab(td);    p.println("side");
        tab(td);    p.println("{");
        
        pl(td+1, "id", String.valueOf(sideid++));       // side id
        pl(td+1, "plane", sv1+" "+sv2+" "+sv3);
        pl(td+1, "material", texture);
        pl(td+1, "uaxis", sua);
        pl(td+1, "vaxis", sva);
        pl(td+1, "rotation","0");
        pl(td+1, "lightmapscale", "16");
        pl(td+1, "smoothing_groups", "0");
        
        tab(td);	p.println("}");
        return surface;
        
    }
    
    public void buildfacemaps()     // build a HashSet array of all faces corresponding to i'th orig face
    {                                  // also calculate the area of ofaces
        System.out.println("Building split face to original face maps");
        origmap = new HashSet[m.ofaces];    // the i'th entry of this array is the set of faces in the i'th oface
        
        for(int i=0;i<m.ofaces;i++)     // create a hashset for each oface
            origmap[i] = new HashSet();
        
        for(int i=0;i<m.faces;i++)      // look at every face
        {
            int o = m.face[i].orig;
            if(o>=0)            // must check for no face correspondence
                origmap[o].add(i);          //  add this face to the set
        }
        
        System.out.println("Building original face areas");
        for(int i=0;i<m.ofaces;i++)     // look at every oface
        {
            Face of = m.oface[i];
            
            Vec [] svec = new Vec[of.numedge];	// list of vectors
            int [] vert = new int[of.numedge];	// list of vertices
            
            int sedge = m.sedge[of.fstedge];
            int v0;					// the start vertex of the face
            if(sedge<0)
                v0 = m.vi2[-sedge];		// backwards wound edge
            else
                v0 = m.vi1[sedge];		// forwards wound edge
            
            float x0 = m.x[v0];         // base vertex coords
            float y0 = m.y[v0];
            float z0 = m.z[v0];
            
            for(int j=0;j<of.numedge;j++)		// build vector list
            {
                sedge = m.sedge[of.fstedge+j];
                // j'th vertex in face
                if(sedge<0)
                    vert[j] = m.vi2[-sedge];
                else
                    vert[j] = m.vi1[sedge];
                
                svec[j] = new Vec(m.x[vert[j]]-x0, m.y[vert[j]]-y0, m.z[vert[j]]-z0);  // the vector from start vertex to i'th
            }
            
            for(int j=1;j<of.numedge-1;j++)
            {
                of.area += 0.5f*svec[j].cross(svec[j+1]).modulus();      // add triangle areas
            }
            if(debug)   System.out.println("OF "+i+" : area "+of.area);
            
            float carea=0.0f;   // area of face components
            Iterator faceit = origmap[i].iterator();    // iterate through the corresponding splitfaces
            while(faceit.hasNext())
                carea += m.face[(Integer)(faceit.next())].area;    // add up the areas of all splitfaces
            
            if(carea > (of.area+area_eps))      // components are bigger, within slop
            {
                of.undersize = true;           // mark the oface
                if(debug) System.out.println("OF "+i+" is undersized: "+carea+">"+of.area);
            }
            
            
            
        }
    }
    
    public void vmfcubemaps()			// write env_cubemaps as entities
    {
        for(int i=0;i<m.cubemaps;i++)
        {
            Cubemap cube = m.cm[i];
            System.out.print(".");
            p.println("entity");
            p.println("{");
            pl(1,"id",String.valueOf(brushid++));
            pl(1,"classname","env_cubemap");
            pl(1,"origin",strvec( new Vec(cube.x,cube.y,cube.z) ));
            pl(1,"cubemapsize",String.valueOf(cube.size));
            
            
            if(envmaps>1)
            {       // count number of sides using this cubemap
                if(cube.sides.size()>envmaps || cube.sides.size()==0)  // too many to write, or none?
                {
                    pl(1,"sides","");       // write blank sidelist
                    
                }
                else
                {   //build side string list
                    
                    Iterator<Integer> sideit = cube.sides.iterator();
                    StringBuffer sstr = new StringBuffer(sideit.next().toString()); // the first sideid
                    
                    while(sideit.hasNext())
                    {
                        sstr.append(" "+sideit.next().toString());      // build list
                    }
                    
                    pl(1,"sides",sstr.toString());      // write list
                }
            }
            else
            {
                pl(1,"sides","");   // output blank list
            }
            
            p.println("}");
        }
    }
    
    
    
    public void vmfstatics()			// write prop_statics as entities
    {
        for(int i=0;i<m.propstatics;i++)
        {
            Pstatic pst = m.ps[i];
            System.out.print(".");
            p.println("entity");
            p.println("{");
            pl(1,"id",String.valueOf(brushid++));
            pl(1,"classname","prop_static");
            pl(1,"origin",strvec( new Vec(pst.ox,pst.oy,pst.oz) ));
            pl(1,"angles",strvec( new Vec(pst.ax,pst.ay,pst.az) ));
            pl(1,"skin",String.valueOf(pst.skin) );
            pl(1,"fademindist",String.valueOf(pst.fademin) );
            pl(1,"fademaxdist",String.valueOf(pst.fademax) );
            pl(1,"solid",String.valueOf(pst.solid) );
            pl(1,"model", m.psname[pst.type] );
            p.println("}");
        }
    }
    
    public void vmfentities()				// write entities, including bmodels
    {
        
        for(int i=1;i<m.entities;i++)		// loop through entities, skip worldspawn (0)
        {
            Entity e = m.ent[i];		// the i'th entity
            
            p.println("entity");
            p.println("{");
            pl(1,"id",String.valueOf(brushid++));
            
            int constart = e.con;				// this is 1st keyval which is connection
            int numkeyvals = e.key.size();
            if(constart<0)	constart = numkeyvals;	// if no connections, emit all keyvals
            
            String skey, sval;
            
            for(int j=0; j<constart; j++)		// emit keyvals
            {
                skey = e.key.get(j);	sval = e.value.get(j);
                if(!skey.equals("model") || e.model<0)
                    pl(1,skey,sval);		// only emit model keyval if not a brush model (model>0)
                
            }
            
            if(numkeyvals>constart)			// connections
            {
                tab(1); p.println("connections"); tab(1); p.println("{");
                
                for(int j=constart; j<numkeyvals; j++)
                {
                    pl(2,e.key.get(j),e.value.get(j));
                }
                tab(1); p.println("}");
            }
            
            if(e.model>0)		// entity has a brush model
            {
                writemodel(e.model, new Vec(e.ox, e.oy, e.oz));
            }
            
            
            p.println("}");
        }
        
    }
    
    public void writeempty(int td, Vec o)       // write an empty model, 16x16x16 cube with invisible texture
    {
        String texture = "TOOLS/TOOLSINVISIBLE";
        
        tab(td); p.println("solid");
        tab(td); p.println("{");
        pl(td+1,"id",String.valueOf(brushid++));	// brush id
        
        for(int i=0;i<3;i++)    // three axes  x/y/z
        {
            for(int j=-1;j<=1;j+=2)  // two directions -/+
            {
                Vec n = new Vec(0,0,0);
                
                n.setcomp(i, (float)j);
                
                float d = 8.0f;                      // distance to origin
                
                Vec e1 = n.scalar(d);                    // e1 is now vector from origin to plane, normal to the surface
                
                Vec udir = new Vec(0,0,1.0f);              // z is z-direction unit vector
                if(i == 2)  // if writing z-normal planes
                {
                    udir = new Vec(1.0f,0,0);  // use x unit-vector as basis
                }   // else use z unit-vector as basis
                
                Vec v1 = udir.cross(n).norm();         // 1st p vector
                // must be perp to normal, so is in plane
                Vec v2 = v1.cross(n).norm();            // 2nd p vector
                // perp to normal and v1, so is also in plane
                v1 = v1.scalar(8.0f);
                v2 = v2.scalar(8.0f);
                
                Vec e2 = e1.add(v1);        // plane vertices
                Vec e3 = e1.add(v2);
                
                e1 = e1.add(o);        // shift due to origin
                e2 = e2.add(o);
                e3 = e3.add(o);
                
                String sv1 = strpvec(e1);
                String sv2 = strpvec(e2);
                String sv3 = strpvec(e3);
                
                Vec u = new Vec(0,0,0);
                Vec v = new Vec(0,0,0);
                
                if(i==0)    // x
                {
                    u.y = 1.0f; v.z = -1.0f;
                }
                else if (i==1)
                {
                    u.x = 1.0f; v.z = -1.0f;
                }
                else if (i==2)
                {
                    u.x = 1.0f; v.y = -1.0f;
                }
                
                String sua = "["+strvec(u)+" 0] 0.25";
                String sva = "["+strvec(v)+" 0] 0.25";
                
                tab(td+1);    p.println("side");
                tab(td+1);    p.println("{");
                
                pl(td+2, "id", String.valueOf(sideid++));       // side id
                pl(td+2, "plane", sv1+" "+sv2+" "+sv3);
                pl(td+2, "material", texture);
                pl(td+2, "uaxis", sua);
                pl(td+2, "vaxis", sva);
                pl(td+2, "rotation","0");
                pl(td+2, "lightmapscale", "12939");
                pl(td+2, "smoothing_groups", "0");
                
                tab(td+1);	p.println("}");
            }
        }
        
        tab(td); p.println("}");        // end of brush
        
    }
    
    public void writemodel(int modelindex, Vec entorg)	// write model of an entity
    {
        Model model = m.mod[modelindex];
        int mfirst = model.fstface;	// 1st world face (0)
        int mnum = model.numface;	// number of world faces
        
        
        if(!fixorg) entorg=new Vec(0,0,0);
        
        
        if(dmode!=3)    // no brushes for this model (wrong mode or not found)
        {
            if(model.numface == 0)      // empty model
                writeempty(1, entorg);
            else
            {
                Face wf;		// current face
                for(int i=0;i<mnum;i++)		// loop through all faces in M0
                {
                    wf = m.face[mfirst+i];	// the i'th face in M0
                    writeflat(1, wf, entorg);	// write this face as flat
                }
            }
        }
        else        // can  write brushes
        {
            for(int i=0;i<model.numbrush;i++)
            {
                Brush b = m.br[model.fstbrush+i];
                writebrush(1, b, entorg);       // write this brush
            }
        }
        
    }
    
    
    
    public void writeflat(int td, Face f, Vec origin)		// write Face f as flat, with td tabs
    {
        // FRONT FACE
        System.out.print(".");
        
        //PLANE VERTICES
        int v1,v2,v3;	// vertices of the plane vectors
        String sv1, sv2, sv3;
        
        if(f.numedge<2) return;		// 0 or 1 edges? Something must be wrong
        
        if(disps && f.dispinfo>-1)   return;     // outputing displacements, and this face is a displacement, so wait for later
        
        if(checkverts)				// new vertex checking?
        {
            Vec [] svec = new Vec[f.numedge];	// list of vectors
            int [] vert = new int[f.numedge];	// list of vertices
            
            int sedge = m.sedge[f.fstedge];
            int v0;					// the start vertex of the face
            if(sedge<0)
                v0 = m.vi2[-sedge];		// backwards wound edge
            else
                v0 = m.vi1[sedge];		// forwards wound edge
            
            float x0 = m.x[v0];
            float y0 = m.y[v0];
            float z0 = m.z[v0];
            
            for(int i=0;i<f.numedge;i++)		// build vector list
            {
                sedge = m.sedge[f.fstedge+i];
                // i'th vertex in face
                if(sedge<0)
                    vert[i] = m.vi2[-sedge];
                else
                    vert[i] = m.vi1[sedge];
                
                svec[i] = new Vec(m.x[vert[i]]-x0, m.y[vert[i]]-y0, m.z[vert[i]]-z0);  // the vector from start vertex to i'th
            }
            
            float maxmcp = -1.0f;	// the largest modulus of cross product found between ixj
            int imax = -1;	// the i index of largest cp
            int jmax = -1;	// the j index of largest cp
            
            for(int i=1;i<f.numedge;i++)					// loop through all i x j combinations
            {
                for(int j=i+1;j<f.numedge;j++)			// ensures j>i
                {
                    float mcp = svec[i].cross(svec[j]).modulus();	//
                    if(mcp>maxmcp)
                    {
                        maxmcp = mcp;
                        imax = i;
                        jmax = j;
                    }
                }
            }
            
            v1 = vert[0];	// 0'th vert is always base vertex
            v2 = vert[imax];	//
            v3 = vert[jmax];	// choose other two such that cross product is maximum
            
        }
        else						// do it the simple way
        {
            int sedge1 = m.sedge[f.fstedge];
            if(sedge1<0)
            {
                v1 = m.vi2[-sedge1];
                v2 = m.vi1[-sedge1];
            }
            else
            {
                v1 = m.vi1[sedge1];
                v2 = m.vi2[sedge1];
            }
            
            int sedge2 = m.sedge[f.fstedge+1];
            
            if(sedge2<0)
            {
                v3 = m.vi1[-sedge2];
            }
            else
            {
                v3 = m.vi2[sedge2];
            }
        }		// end of vertex chooser
        
        Vec e1, e2, e3;		// shift vertices by origin vector
        // note entity model writer controls origin input
        e1 = new Vec(m.x[v1],m.y[v1],m.z[v1]).add(origin);
        e2 = new Vec(m.x[v2],m.y[v2],m.z[v2]).add(origin);
        e3 = new Vec(m.x[v3],m.y[v3],m.z[v3]).add(origin);
        // *** round vertex coordinates here
        
        Vec ev12 = e2.subtract(e1);
        Vec ev13 = e3.subtract(e1);
        Vec norm = ev12.cross(ev13).norm();
        
        if(norm.notvalid())     // bad plane normal
        {
            if(debug) System.out.print("\nBad normal! "+ev12+"x"+ev13);
            return;
        }
        
        sv1 = strpvec(e1);       //
        sv2 = strpvec(e2);
        sv3 = strpvec(e3);
        
        
        //FACE TEXTURE
        
        Texture tex = m.gettex(f.texinfo, origin, fp);
        
        String texture = "NULL";
        if(ttexmode==0)	texture = "SKYBOX/SKYFAKEWHITE";
        if(ttexmode==1) texture = "TOOLS/TOOLSBLACK";
        if(ttexmode==2)	texture = "DEV/DEV_MEASUREGENERIC01";
        if(ttexmode==3)	texture = tex.name;
        
        //check if texture used envmap
        
        if(envmaps>1)       // if outputting cubemaps with side lists
        {
            int envmapid = tex.envmap;
            if(envmapid!=-1) // if a cubemap is used for this texture
                m.cm[envmapid].sides.add(new Integer(sideid));         // add the sideid to the cubemap list
            
        }
        
        String sua = tex.uaxis;
        String sva = tex.vaxis;
        
        // WRITE THE FRONT FACE
        tab(td); p.println("solid");
        tab(td); p.println("{");
        
        pl(td+1,"id",String.valueOf(brushid++));	// brush id
        
        tab(td+1);	p.println("side");
        tab(td+1);	p.println("{");
        
        pl(td+2, "id", String.valueOf(sideid++));	// side id
        pl(td+2, "plane", sv1+" "+sv2+" "+sv3);
        pl(td+2, "material", texture);
        pl(td+2, "uaxis", sua);
        pl(td+2, "vaxis", sva);
        pl(td+2, "rotation","0");
        pl(td+2, "lightmapscale", "16");
        pl(td+2, "smoothing_groups", String.valueOf(f.smooth));
        
        tab(td+1);	p.println("}");
        // end of front face
        
        if(stexmode==0)	texture = "SKYBOX/SKYFAKEWHITE";
        if(stexmode==1) texture = "TOOLS/TOOLSBLACK";
        if(stexmode==2)	texture = "TOOLS/TOOLSINVISIBLE";
        // ==3, texture same as front face
        
        
        
/*        if(v1==v2 || v2==v3 || v3==v1)		// degenerate vertices
        {
            norm = new Vec(1,0,0);			// would give bad normals, so make one up
        } */
        
        
        switch(bmode)
        {
            case 0:		// pyrimidal
                writepyramback(f, td, norm, origin, texture, sua, sva);
                break;
            case 1:		// prismatic
                writeprismback(f, td, norm, origin, texture);
                break;
            case 2:		// reciprocal face
                writeflatback(f, td, origin, texture, sua, sva);
                break;
        }
        
        // end of brush
        tab(td);	p.println("}");
        
    }
    
    public void writeflatback(Face f, int td, Vec origin, String tex, String sua, String sva)
    {							// write back of Face f
        // this doesn't work
        // BACK FACE
        
        //PLANE VERTICES
        int v1,v2,v3;	// vertices of the plane vectors
        String sv1, sv2, sv3;
        
        if(f.numedge<2) return;		// 0 or 1 edges? Something must be wrong
        
        int sedge1 = m.sedge[f.fstedge];
        if(sedge1<0)
        {
            v1 = m.vi2[-sedge1];
            v2 = m.vi1[-sedge1];
        }
        else
        {
            v1 = m.vi1[sedge1];
            v2 = m.vi2[sedge1];
        }
        
        int sedge2 = m.sedge[f.fstedge+1];
        
        if(sedge2<0)
        {
            v3 = m.vi1[-sedge2];
        }
        else
        {
            v3 = m.vi2[sedge2];
        }
        
        Vec e1, e2, e3;		// shift vertices by origin vector
        // note entity model writer controls origin input
        e1 = new Vec(m.x[v1],m.y[v1],m.z[v1]).add(origin);
        e2 = new Vec(m.x[v2],m.y[v2],m.z[v2]).add(origin);
        e3 = new Vec(m.x[v3],m.y[v3],m.z[v3]).add(origin);
        // *** round vertex coordinates here
        
        sv1 = strpvec(e1);       //
        sv2 = strpvec(e2);
        sv3 = strpvec(e3);
        
        
        
        
        tab(td+1);	p.println("side");
        tab(td+1);	p.println("{");
        
        pl(td+2, "id", String.valueOf(sideid++));	// side id
        pl(td+2, "plane", sv1+" "+sv2+" "+sv3);
        pl(td+2, "material", tex);
        pl(td+2, "uaxis", sua);
        pl(td+2, "vaxis", sva);
        pl(td+2, "rotation","0");
        pl(td+2, "lightmapscale", "16");
        pl(td+2, "smoothing_groups", String.valueOf(f.smooth));
        
        tab(td+1);	p.println("}");
        // end of front face
        
    }
    
    
    public void writepyramback(Face f, int t, Vec norm, Vec o, String tex, String sua, String sva) // write the back faces of a pyramid brush
    {
        float nx=0, ny=0, nz=0;		// the coords of the pyramid vertex
        
        // find face barycentre
        
        
        
        for(int i=0; i<f.numedge;i++)
        {
            int se = m.sedge[f.fstedge+i];
            int vertex;
            if(se<0)			// find the unique vertex on the edge
                vertex = m.vi2[-se];
            else
                vertex = m.vi1[se];
            
            nx += m.x[vertex];
            ny += m.y[vertex];
            nz += m.z[vertex];
        }
        nx /= f.numedge;			// average vertex position
        ny /= f.numedge;
        nz /= f.numedge;
        
        nx += norm.x * depth; 		// displace vertex from face in normal direction by depth
        ny += norm.y * depth;
        nz += norm.z * depth;
        
        // *** round vertex coordinates here
        
        /*if(stexmode!=3)
        {
            sua = "[1 0 0 0] 0.25";
            sva = "[0 0 1 0] 0.25";
        }*/
        
        int v1, v2;		// edge vertices
        
        for(int i=0; i<f.numedge;i++)
        {
            int se = m.sedge[f.fstedge+i];
            if(se<0)
            {
                v1 = m.vi2[-se];
                v2 = m.vi1[-se];
            }
            else
            {
                v1 = m.vi1[se];
                v2 = m.vi2[se];
            }
            
            Vec e1 = new Vec(m.x[v1],m.y[v1],m.z[v1]).add(o);	// vertices of the side face
            Vec e2 = new Vec(m.x[v2],m.y[v2],m.z[v2]).add(o);
            Vec e3 = new Vec(nx, ny, nz).add(o);
            writebackface(t, e1, e2, e3, tex, sua, sva);
        }
        
    }
    
    public void writeprismback(Face f, int t, Vec norm, Vec o, String tex) // write the back and side faces of a slab
    {
        
    }
    
    public void writebackface(int td, Vec v1, Vec v2, Vec v3, String tex, String sua, String sva)  // write the side face of a pyramid
    {
        tab(td+1);	p.println("side");
        tab(td+1);	p.println("{");
        String sv1 = strpvec(v1);    // swap normal
        String sv2 = strpvec(v3);
        String sv3 = strpvec(v2);
        
        pl(td+2, "id", String.valueOf(sideid++));	// side id
        pl(td+2, "plane", sv1+" "+sv2+" "+sv3);
        pl(td+2, "material", tex);
        pl(td+2, "uaxis", sua);
        pl(td+2, "vaxis", sva);
        pl(td+2, "rotation","0");
        pl(td+2, "lightmapscale", "16");
        pl(td+2, "smoothing_groups", "0");
        
        tab(td+1);	p.println("}");
        
    }
    
    public void vmforigfaces()
    {
        Model M0 = m.mod[0];		// Model 0 = world brushes
        int wfirst = M0.fstface;	// 1st world face (0)
        int wnum = M0.numface;	// number of world faces
        
        Face wf;		// current face
        Face owf;		// current original face
        for(int i=0;i<wnum;i++)		// loop through all faces in M0
        {
            wf = m.face[wfirst+i];	// the i'th face in M0
            if(wf.orig>=0)
            {
                owf = m.oface[wf.orig];	// the original face corresponding to this
                if(!owf.written)		// has it been written yet?
                {
                    writeflat(1, owf, org);	// write this oface as flat
                    owf.written=true;			// mark as written
                }
            }
        }
        if(!disps) p.println("}");	// end of world brushes (only if no disps)
        
    }
    
    public void vmforigfacesplus()          // write orig faces, unless one is undersized, in which case write its split faces
    {
        Model M0 = m.mod[0];		// Model 0 = world brushes
        int wfirst = M0.fstface;	// 1st world face (0)
        int wnum = M0.numface;	// number of world faces
        
        Face wf;		// current face
        Face owf;		// current original face
        for(int i=0;i<wnum;i++)		// loop through all faces in M0
        {
            wf = m.face[wfirst+i];	// the i'th face in M0
            if(wf.orig>=0)
            {
                owf = m.oface[wf.orig];	// the original face corresponding to this
                if(!owf.written)		// has it been written yet?
                {
                    if(owf.undersize)       // oface is undersized!
                    {                       // write it as split faces
                        oucount++;
                        Iterator fit = origmap[wf.orig].iterator();
                        if(debug) System.out.print("\n OF "+wf.orig+": ");
                        while(fit.hasNext())                        // iterate through the corresponding faces
                        {
                            int findex = (Integer)fit.next();
                            writeflat(1, m.face[findex], org); // write them
                            if(debug) System.out.print(findex+" ");
                        }
                        owf.written=true;
                    }
                    else                // oface is not undersized
                    {
                        writeflat(1, owf, org);	// write this oface as flat
                        owf.written=true;			// mark as written
                    } // undersize
                } //written
            } // o exists
            else                            // no corresponding oface
            {
                writeflat(1, wf, org);      // write the face directly
            }
        }
        if(!disps) p.println("}");	// end of world brushes (only if no disps)
        
    }
    
    
    
    public void vmffaces()			// write by faces
    {
        Model M0 = m.mod[0];		// Model 0 = world brushes
        int wfirst = M0.fstface;	// 1st world face (0)
        int wnum = M0.numface;	// number of world faces
        
        Face wf;		// current face
        for(int i=0;i<wnum;i++)		// loop through all faces in M0
        {
            wf = m.face[wfirst+i];	// the i'th face in M0
            writeflat(1, wf, org);	// write this face as flat
        }
        
        if(!disps) p.println("}");	// end of world brushes (only if no disps)
        
        
    }
    
    public void vmfworldspawn()		// write worldspawn header
    {
        Entity ws = m.ent[0];
        p.println("world"); p.println("{");
        pl(1,"id","1");
        pl(1,"mapversion","2");
        String skey, sval;
        Iterator<String> kit = ws.key.iterator();
        Iterator<String> vit = ws.value.iterator();
        
        while(kit.hasNext())				// iterate through ws's k/vs
        {
            skey = kit.next();
            sval = vit.next();
            
            if(!skey.equals("world_maxs")		// k/vs to exclude
            && !skey.equals("world_mins") )
                pl(1,skey,sval);		// write remaining k/vs
        } //while
        pl(1,"comment", "Decompiled by VMEX "+VERSION+" from "+filename);
    }
    
    public void vmfheader()			// write the vmf header
    {
        p.println("versioninfo");
        p.println("{");
        
        pl(1,"editorvesion","400");
        pl(1,"editorbuild","2978");
        pl(1,"mapversion","2");
        pl(1,"formatversion","100");
        pl(1,"prefab","0");
        
        p.println("}");
        
        p.println("visgroups");
        p.println("{");
        tab(1); p.println("visgroup");
        tab(1); p.println("{");
        pl(2,"name", "Auto");
        pl(2,"visgroupid","1");
        pl(2,"color", "152 185 142");
        tab(2); p.println("visgroup");
        tab(2); p.println("{");
        pl(3,"name", "Entities");
        pl(3,"visgroupid", "3");
        pl(3,"color", "104 137 238");
        tab(2);  p.println("}");
        
        tab(2); p.println("visgroup");
        tab(2); p.println("{");
        pl(3,"name", "Detail Brushes");
        pl(3,"visgroupid", "2");
        pl(3,"color", "240 25 150");
        tab(2);  p.println("}");
        
        tab(2); p.println("visgroup");
        tab(2); p.println("{");
        pl(3,"name", "Brush Entities");
        pl(3,"visgroupid", "4");
        pl(3,"color", "231 148 165");
        tab(2);  p.println("}");
        
        tab(2); p.println("visgroup");
        tab(2); p.println("{");
        pl(3,"name", "Clip Brushes");
        pl(3,"visgroupid", "6");
        pl(3,"color", "255 128 128");
        tab(2);  p.println("}");
        
        tab(2); p.println("visgroup");
        tab(2); p.println("{");
        pl(3,"name", "Hint Brushes");
        pl(3,"visgroupid", "7");
        pl(3,"color", "100 120 190");
        tab(2);  p.println("}");
        tab(1);  p.println("}");
        p.println("}");
        
        p.println("viewsettings");
        p.println("{");
        
        pl(1,"bSnapToGrid","0");
        pl(1,"bShowGrid","1");
        pl(1,"nGridSpacing","128");
        pl(1,"bShow3DGrid","0");
        
        p.println("}");
        
    }
    
    public void vmftrailer()
    {
        
        p.println("cameras");
        p.println("{");
        
        pl(1,"activecamera","-1");
        
        p.println("}");
        p.println("cordon");
        p.println("{");
        
        pl(1,"mins","(99999 99999 99999)");
        pl(1,"maxs","(-99999 -99999 -99999)");
        pl(1,"active","0");
        
        p.println("}");
        
    }
    
    public String strpvec(Vec v)		// return string constructed from vector "(xxx yyy zzz)"
    {
        return "("+strvec(v)+")";
    }
    
    public String strvec(Vec v)		// return string constructed from vector "xxx yyy zzz"
    {
        return fp.format(v.x)+" "+fp.format(v.y)+" "+fp.format(v.z);
    }
    
    
    public void pl(int tabdepth, String a, String b)	// emit quoted strings at given tabdepth
    {
        tab(tabdepth);
        p.println("\""+a+"\" \""+b+"\"");
    }
    
    public void tab(int depth)				// emit depth tabs
    {
        for(int i=0;i<depth;i++)	p.print("\t");
    }
    
    public void fixtexturenames()			// chop "maps/mapname/[maps/mapname/]" from start of names
    {                                                      // chop "_n_n_n or _n_n_n_depth_n from ends of names
        if(fixtex) System.out.println("Fixing envmapped texture names");
        for(int i=0;i<m.texname.length;i++)
        {
            String tex = m.texname[i];
            if(debug)   System.out.println("Fixup: "+tex);
            if(tex.equalsIgnoreCase("tools/locked"))    m.no_decomp = true;
            if(!fixtex) continue;
            
            String [] token = tex.split("/");
            int tokens = token.length;
            if(token[0].equalsIgnoreCase("maps"))
            {
                int cx=0, cy=0, cz=0;   // cubemap origin tied to this face
                
                String ftex = token[tokens-1]; // tex_0_0_0 or tex_0_0_0_depth_0 or tex_depth_0
                
                
                String [] utex = ftex.split("_");       // split at underscores
                int utokens = utex.length;
                int endtoken=utokens;
                
                boolean deporg = Pattern.matches(".+_-?\\d+_-?\\d+_-?\\d+_depth_-?\\d+$",ftex);   // matches x_0_0_0_depth_0
                boolean deponly = Pattern.matches(".+_depth_-?\\d+$",ftex);                       // matches x_depth_0 or previous
                boolean orgonly = Pattern.matches(".+_-?\\d+_-?\\d+_-?\\d+$",ftex);               // matches x_0_0_0
                
                if(debug) System.out.print("do:"+deporg+" d:"+deponly+" o:"+orgonly);
                
                if(deponly)     // is of form tex_depth_0
                {
                    endtoken = utokens-2;       // truncate
                }
                
                
                if(deporg)     // is of form tex_0_0_0_depth_0
                {
                    cx = Integer.parseInt(utex[utokens-5]);
                    cy = Integer.parseInt(utex[utokens-4]);
                    cz = Integer.parseInt(utex[utokens-3]);
                    
                    endtoken = utokens-5;   // the last actual texture name token
                    
                }
                if(orgonly)
                {                                    // is of form tex_0_0_0
                    cx = Integer.parseInt(utex[utokens-3]);
                    cy = Integer.parseInt(utex[utokens-2]);
                    cz = Integer.parseInt(utex[utokens-1]);
                    
                    endtoken = utokens-3;   // the last actual texture name token
                }
                
                if(!deponly && !deporg && !orgonly)
                    System.out.println("Unknown texture type: "+ftex);
                
                
                StringBuffer otex = new StringBuffer(utex[0]);
                
                for(int j=1;j<endtoken;j++)
                    otex.append("_"+utex[j]);       // rebuild name from underscore tokens
                
                if(debug)   System.out.println(" -> "+otex.toString());
                
                if(envmaps>1 && !deponly)       // outputting cubemap side lists and found an envmapped tex
                {
                    for(int j=0;j<m.cubemaps;j++)     // scan all cubemaps
                    {
                        if(cx==m.cm[j].x && cy==m.cm[j].y && cz==m.cm[j].z)    // check for match
                        {
                            
                            m.texenv[i] = j;       // set cubemap index used by this texdata/texname
                            
                            break;
                        }
                    }
                    
                }
                
                
                
                
                StringBuffer outtex = new StringBuffer();
                if(tokens>2)
                {
                    int stok = 2;			// is of form /maps/name/texname
                    if(token[2].equalsIgnoreCase("maps"))	// is of form /maps/name/maps/name/texname
                    {
                        stok = 4;
                    }
                    
                    for(int j=stok;j<tokens-1;j++)
                        outtex.append(token[j]).append("/");
                    
                    outtex.append(otex);
                    
                }
                m.texname[i]=outtex.toString();
                
            }
            
        }
        
        
    }
    
    public void protbrushcheck()
    {
        if(debug) System.out.println("protbrush");
        
        boolean m1=false, m2=false, m3=false;
        float Eps_size = 0.01f;
        
        for(int i=0;i<m.brushes;i++)     // check every brush
        {
            
            Brush b = m.br[i];
            
            if(alignedbrush(b))
            {
                
                if(sametexbrush(b))
                {
                    
                    Vec bsize = getsize(i);
                    if( new Vec(1,4,9).subtract(bsize).modulus() < Eps_size)
                        m1 = true;
                    if( new Vec(4,9,1).subtract(bsize).modulus() < Eps_size)
                        m2 = true;
                    if( new Vec(9,1,4).subtract(bsize).modulus() < Eps_size)
                        m3 = true;
                }
            }
            
        }
        
        if(m1 && m2 && m3)  m.no_decomp = true;
    }
    
    public Vec getsize(int brindex)   // returns the x, y and z extents of a brush, as a vector
    {
        
        Vec mins = new Vec(Wind.Maxlen, Wind.Maxlen, Wind.Maxlen);
        Vec maxs = new Vec(-Wind.Maxlen, -Wind.Maxlen, -Wind.Maxlen);
        
        Brush brush = m.br[brindex];
        
        for(int i=0;i<brush.numside;i++)
        {
            Wind wside = Wind.windfromside(m, brush, i);
            for(int j=0;j<wside.size();j++)
            {
                Vec v = wside.v.get(j);
                mins = v.minbound(mins);
                maxs = v.maxbound(maxs);
            }
        }
        return maxs.subtract(mins);
    }
    
    public boolean alignedbrush(Brush brush)
    {
        
        float Aligned_alpha = 1.0f - 0.01f;
        
        if(brush.numside != 6)  return false;
        
        for(int i=0;i<6;i++)
        {
            Plane bpl = m.pl[m.bsd[brush.fstside+i].pnum];
            Vec pnorm = bpl.getvec();
            
            boolean axis = false;
            for(int j=0;j<3;j++)
            {
                
                if(Math.abs(pnorm.comp(j)) > Aligned_alpha)
                    axis = true;
            }
            
            if(!axis)   return false;
        }
        
        return true;
    }
    
    public boolean sametexbrush(Brush brush)
    {
        //// -1 texinfo problem!
        Brushside brsd = m.bsd[brush.fstside];
        Texture tex = m.gettex(brsd.texinfo);
        String texname = tex.name;
        
        if(tex==null) return false;
        boolean pass = true;
        
        for(int i=1;i<brush.numside;i++)
        {
            brsd = m.bsd[brush.fstside+i];
            String nexttexname = m.gettex(brsd.texinfo).name;
            if(!texname.equalsIgnoreCase(nexttexname))
                pass = false;
        }
        return pass;
    }
    
    
    
    public void args(String args[])		// parse the command line arguments
    {
        if(args.length ==0)
        {
            usage();
            System.exit(0);
        }
        
        try
        {
            
            for(int i=0;i<args.length-1;i++)
            {
                //System.out.print(args[i]+"  ");
                if(args[i].length()>1)
                {
                    if(args[i].charAt(0)=='-')
                    {
                        char c = args[i].charAt(1);
                        
                        switch(c)
                        {
                            case 'x':
                                fixtex = false;
                                break;
                            case 'p':
                                props = false;
                                break;
                            case 'o':
                                fixorg = false;
                                break;
                            case 'd':
                                depth = Float.parseFloat(args[i].substring(2));
                                break;
                            case 'r':
                                prec = Integer.parseInt(args[i].substring(2));
                                if(prec<0)
                                    prec=0;
                                break;
                            case 'c':
                                envmaps = Integer.parseInt(args[i].substring(2));
                                if(envmaps<0)
                                    envmaps=0;
                                break;
                            case 's':
                                stexmode = Integer.parseInt(args[i].substring(2));
                                if(stexmode<0 || stexmode>3)
                                    stexmode = 2;
                                break;
                            case 't':
                                ttexmode = Integer.parseInt(args[i].substring(2));
                                if(ttexmode<0 || ttexmode>3)
                                    ttexmode = 3;
                                break;
                            case 'D':
                                debug = true;
                                System.out.println("Debugging output on");
                                break;
                            case 'm':
                                dmode = Integer.parseInt(args[i].substring(2));
                                if(dmode<0 || dmode>2)
                                    dmode = 2;
                                break;
                            case 'v':
                                checkverts = false;
                                break;
                            case 'h':
                                disps = false;
                                break;
                                
                            default:
                                System.out.println("Unrecognised argument "+args[i]);
                                usage();
                                System.exit(0);
                        }
                        
                    }
                    else
                    {
                        System.out.println("Unrecognised argument "+args[i]);
                        usage();
                        System.exit(0);
                    }
                }
            }
        }
        catch(NumberFormatException nfe)
        {
            System.out.println(nfe);
            usage();
            System.exit(0);
        }
        filename = args[args.length-1];
        
    }
    
    public void usage()
    {
        System.out.println("Usage: VMEX [options] filename\n" +
                "where option are:\n" +
                "  -m<mode>\tDecompiling mode: 0 original faces, 1 split faces,\n" +
                "  \t\t2 orig and split faces, 3 brushes and planes (default)\n" +
                "  -d<value>\tThickness of created brushes (default 1.0)\n" +
                "  -t<mode>\tFace texture: 0 white, 1 black, 2 orange, 3 original (default)\n" +
                "  -s<mode>\tBack texture: 0 white, 1 black, 2 invis (default), 3 as face\n" +
                "  -r<value>\tDecimal precision of output floating point values (default 8)\n" +
                "  -o\t\tDon't fixup origins of brush models\n" +
                "  -p\t\tDon't output prop_statics\n" +
                "  -x\t\tDon't fixup envmapped texture names\n" +
                "  -c<value>\tCubemaps: 0 none, 1 origin only, n Up to n sides (default 8)\n"+
                "  -v\t\tDon't do optimised vertex picking\n" +
                "  -h\t\tDon't output displacement surfaces\n\n"
                
                );
        
        
    }
    
    public static void main(String args[])
    {
        Vmex inst = new Vmex();		// create an instance
        System.out.println("Valve Map Extractor "+VERSION+" - by Rof (rof@mellish.org.uk)");
        inst.args(args);		// parse arguments
        inst.exec();			// execute
    }
}
